"""
Handle HTTP download requests from content URLs
"""

import os
import shutil
import zipfile

from tqdm import tqdm
from urllib.request import urlopen
from construe.exceptions import DownloadError

from .signature import sha256sum


# Download chunk size
CHUNK = 524288


def download_zip(url, out, signature, replace=False, extract=True, progress=True):
    """
    Download a zipped file at the given URL saving it to the out directory. Once
    downloaded, verify the signature to make sure the download hasn't been tampered
    with or corrupted. If the file already exists it will be overwritten only if
    replace=True. If extract=True then the file will be unzipped.
    """
    # Get the name of the file from the URL
    basename = os.path.basename(url)
    name, _ = os.path.splitext(basename)

    # Get the archive and data directory paths
    archive = os.path.join(out, basename)
    datadir = os.path.join(out, name)

    # If the archive exists cleanup or raise override exception
    if os.path.exists(archive):
        if not replace:
            raise DownloadError(
                f"file already exists at {archive}, set replace=False to overwrite"
            )

        shutil.rmtree(datadir)
        os.remove(archive)

    # Create the output directory if it does not exist
    if not os.path.exists(datadir):
        os.mkdir(datadir)

    # Fetch the response in a streaming fashion and write it to disk.
    response = urlopen(url)
    content_length = int(response.headers["Content-Length"])

    pbar = None
    if progress:
        pbar = tqdm(
            unit="B", total=content_length, desc=f"Downloading {basename}", leave=False
        )

    with open(archive, "wb") as f:
        while True:
            chunk = response.read(CHUNK)
            if not chunk:
                break
            f.write(chunk)

            if pbar:
                pbar.update(len(chunk))

    # Compare the signature of the archive to the expected one
    if sha256sum(archive) != signature:
        raise DownloadError("Download signature does not match hardcoded signature!")

    # If extract, extract the zipfile.
    if extract:
        zf = zipfile.ZipFile(archive)
        zf.extractall(path=datadir)



"""
Primary entry point for construe CLI application
"""

import click
import platform

from datetime import datetime

from .version import get_version
from .utils import resolve_exclude
from .exceptions import DeviceError

from .datasets.path import get_data_home
from .datasets.loaders import cleanup_all_datasets
from .datasets.download import (
    download_all_datasets,
    download_dialects,
    download_lowlight,
    download_reddit,
    download_movies,
    download_essays,
    download_aegis,
    download_nsfw,
)

from .models.path import get_model_home
from .models.loaders import cleanup_all_models
from .models.download import (
    download_all_models,
    download_moondream,
    download_whisper,
    download_mobilenet,
    download_mobilevit,
    download_nsfw as download_nsfw_model,
    download_offensive,
    download_lowlight as download_lowlight_model,
    download_gliner,
)


from .nsfw import NSFW
from .gliner import GLiNER
from .whisper import Whisper
from .lowlight import LowLight
from .mobilenet import MobileNet
from .mobilevit import MobileViT
from .moondream import MoonDream
from .offensive import Offensive
from .basic import BasicBenchmark

from .benchmark import BenchmarkRunner


CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
}

DATASETS = [
    "all",
    "dialects",
    "lowlight",
    "reddit",
    "movies",
    "essays",
    "aegis",
    "nsfw",
]

MODELS = [
    "all",
    "moondream",
    "whisper",
    "mobilenet",
    "mobilevit",
    "nsfw",
    "offensive",
    "lowlight",
    "gliner",
]

BENCHMARKS = {
    "whisper": Whisper,
    "lowlight": LowLight,
}


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(get_version(), message="%(prog)s v%(version)s")
@click.option(
    "-o",
    "--out",
    default=None,
    type=str,
    help="specify the path to write the benchmark results to",
)
@click.option(
    "-d",
    "--device",
    default=None,
    type=str,
    envvar=["CONSTRUE_DEVICE", "TORCH_DEVICE"],
    help="specify the pytorch device to run on e.g. cpu, mps or cuda",
)
@click.option(
    "-e",
    "--env",
    default=None,
    envvar=["CONSTRUE_ENV", "ENV"],
    help="name of the experimental environment for comparison (default is hostname)",
)
@click.option(
    "-c",
    "--count",
    default=1,
    type=int,
    help="specify the number of times to run each benchmark",
)
@click.option(
    "-l",
    "--limit",
    default=None,
    type=int,
    help="limit the number of instances to inference on in each benchmark",
)
@click.option(
    "-D",
    "--datadir",
    default=None,
    envvar="CONSTRUE_DATA",
    help="specify the location to download datasets to",
)
@click.option(
    "-M",
    "--modeldir",
    default=None,
    envvar="CONSTRUE_MODELS",
    help="specify the location to download models to",
)
@click.option(
    "-S",
    "--sample/--no-sample",
    default=True,
    help="use sample dataset instead of full dataset for benchmark",
)
@click.option(
    "-C",
    "--cleanup/--no-cleanup",
    default=True,
    help="cleanup all downloaded datasets after the benchmark is run",
)
@click.option(
    "-Q",
    "--verbose/--quiet",
    default=True,
    help="specify the verbosity of the output and progress bars",
)
@click.pass_context
def main(
    ctx,
    out=None,
    env=None,
    device=None,
    count=1,
    limit=None,
    datadir=None,
    modeldir=None,
    sample=True,
    cleanup=True,
    verbose=True,
):
    """
    A utility for executing inferencing benchmarks.
    """
    if device is not None:
        try:
            import torch
            torch.set_default_device(device)
        except RuntimeError as e:
            raise DeviceError(e)

        click.echo(f'using torch.device("{device}")')

    if env is None:
        env = platform.node()

    if out is None:
        out = f"construe-results-{datetime.now().strftime('%Y%m%d%H%M%S')}.json"

    ctx.ensure_object(dict)
    ctx.obj["out"] = out
    ctx.obj["device"] = device
    ctx.obj["env"] = env
    ctx.obj["n_runs"] = count
    ctx.obj["limit"] = limit
    ctx.obj["data_home"] = get_data_home(datadir)
    ctx.obj["model_home"] = get_model_home(modeldir)
    ctx.obj["use_sample"] = sample
    ctx.obj["cleanup"] = cleanup
    ctx.obj["verbose"] = verbose


@main.command()
@click.option(
    "-E",
    "--exclude",
    default=None,
    type=click.Choice(BENCHMARKS.keys(), case_sensitive=False),
    help="specify benchmarks to exclude from runner",
)
@click.option(
    "-I",
    "--include",
    default=None,
    type=click.Choice(BENCHMARKS.keys(), case_sensitive=False),
    help="specify benchmarks to include in runner",
)
@click.pass_context
def run(ctx, **kwargs):
    """
    Executes all available benchmarks.
    """
    out = ctx.obj.pop("out")
    exclude = resolve_exclude(
        kwargs.pop("exclude"), kwargs.pop("include"), BENCHMARKS.keys()
    )
    benchmarks = [bench for name, bench in BENCHMARKS.items() if name not in exclude]

    runner = BenchmarkRunner(benchmarks=benchmarks, **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.option(
    "-o",
    "--saveto",
    default=None,
    help="path to write the measurements pickle data to",
)
@click.option(
    "-t",
    "--num-threads",
    default=None,
    type=int,
    help="specify number of threads for benchmark (default to maximum)",
)
@click.option(
    "-F",
    "--fuzz/--no-fuzz",
    default=False,
    help="fuzz the tensor sizes of the inputs to the benchmark",
)
@click.option(
    "-S",
    "--seed",
    default=None,
    type=int,
    help="set the random seed for random generation",
)
@click.pass_context
def basic(ctx, **kwargs):
    """
    Runs basic dot product performance benchmarks.
    """
    kwargs["env"] = ctx.obj["env"]
    if kwargs["saveto"] is None:
        kwargs["saveto"] = ctx.obj["out"]

    benchmark = BasicBenchmark(**kwargs)
    benchmark.run()


@main.command()
@click.pass_context
def moondream(ctx, **kwargs):
    """
    Executes image-to-text inferencing benchmarks.
    """
    out = ctx.obj.pop("out")
    runner = BenchmarkRunner(benchmarks=[MoonDream], **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.pass_context
def whisper(ctx, **kwargs):
    """
    Executes audio-to-text inferencing benchmarks.
    """
    out = ctx.obj.pop("out")
    runner = BenchmarkRunner(benchmarks=[Whisper], **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.pass_context
def mobilenet(ctx, **kwargs):
    """
    Executes image classification inferencing benchmarks.
    """
    out = ctx.obj.pop("out")
    runner = BenchmarkRunner(benchmarks=[MobileNet], **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.pass_context
def mobilevit(ctx, **kwargs):
    """
    Executes object detection inferencing benchmarks.
    """
    out = ctx.obj.pop("out")
    runner = BenchmarkRunner(benchmarks=[MobileViT], **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.pass_context
def nsfw(ctx, **kwargs):
    """
    Executes NSFW image classification inferencing benchmarks.
    """
    out = ctx.obj.pop("out")
    runner = BenchmarkRunner(benchmarks=[NSFW], **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.pass_context
def lowlight(ctx, **kwargs):
    """
    Executes lowlight image enhancement inferencing benchmarks.
    """
    out = ctx.obj.pop("out")
    runner = BenchmarkRunner(benchmarks=[LowLight], **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.pass_context
def offensive(ctx, **kwargs):
    """
    Executes offensive speech text classification inferencing benchmarks.
    """
    out = ctx.obj.pop("out")
    runner = BenchmarkRunner(benchmarks=[Offensive], **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.pass_context
def gliner(ctx, **kwargs):
    """
    Executes GLiNER named entity discovery inferencing benchmarks.
    """
    out = ctx.obj.pop("out")
    runner = BenchmarkRunner(benchmarks=[GLiNER], **ctx.obj)
    runner.run()
    runner.save(out)


@main.command()
@click.option(
    "-C",
    "--clean/--no-clean",
    default=False,
    help="cleanup the downloaded data cache and exit",
)
@click.option(
    "-d",
    "--download",
    default=None,
    type=click.Choice(DATASETS, case_sensitive=False),
)
@click.option(
    "-S",
    "--sample/--no-sample",
    default=True,
    help="if downloading a dataset, download only a sample",
)
@click.option(
    "-v",
    "--verbose/--no-verbose",
    default=False,
    help="print verbose info, otherwise just print the datasets path",
)
@click.pass_context
def datasets(ctx, clean=False, download=None, sample=True, verbose=False):
    """
    Helper utility for managing the dataset cache.
    """
    if clean and download:
        raise click.ClickException("cannot specify both --clean and --download")

    data_home = ctx.obj["data_home"]
    if clean:
        cleanup_all_datasets(data_home)
        return

    if download is not None:
        downloader = {
            "all": download_all_datasets,
            "dialects": download_dialects,
            "lowlight": download_lowlight,
            "reddit": download_reddit,
            "movies": download_movies,
            "essays": download_essays,
            "aegis": download_aegis,
            "nsfw": download_nsfw,
        }[download]

        downloader(sample=sample, data_home=data_home, progress=True)
        return

    # Provide some info
    if verbose:
        print(f"downloaded datasets are stored in {data_home}")
    else:
        print(data_home)


@main.command()
@click.option(
    "-C",
    "--clean/--no-clean",
    default=False,
    help="cleanup the downloaded models cache and exit",
)
@click.option(
    "-d",
    "--download",
    default=None,
    type=click.Choice(MODELS, case_sensitive=False),
)
@click.option(
    "-v",
    "--verbose/--no-verbose",
    default=False,
    help="print verbose info, otherwise just print the models path",
)
@click.pass_context
def models(ctx, clean=False, download=None, sample=True, verbose=False):
    """
    Helper utility for managing the models cache.
    """
    if clean and download:
        raise click.ClickException("cannot specify both --clean and --download")

    model_home = ctx.obj["model_home"]
    if clean:
        cleanup_all_models(model_home)
        return

    if download is not None:
        downloader = {
            "all": download_all_models,
            "moondream": download_moondream,
            "whisper": download_whisper,
            "mobilenet": download_mobilenet,
            "mobilevit": download_mobilevit,
            "nsfw": download_nsfw_model,
            "offensive": download_offensive,
            "lowlight": download_lowlight_model,
            "gliner": download_gliner,
        }[download]

        downloader(model_home=model_home, progress=True)
        return

    # Provide some info
    if verbose:
        print(f"downloaded models are stored in {model_home}")
    else:
        print(model_home)


if __name__ == "__main__":
    main(
        obj={},
        prog_name="construe",
    )



"""
Upload models to the google cloud bucket (developers only).
"""

import os
import glob
import json

from ..exceptions import UploadError

try:
    from google.cloud import storage
except ImportError:
    storage = None


CONSTRUE_BUCKET = "construe"
GOOGLE_CREDENTIALS = "GOOGLE_APPLICATION_CREDENTIALS"


def upload(name, path, client=None, bucket=CONSTRUE_BUCKET):
    """
    Upload data from source path to a bucket with destination name.
    """
    if client is None:
        client = connect_storage()

    if not os.path.exists(path) or not os.path.isfile(path):
        raise UploadError("no zip file exists at " + path)

    try:
        bucket = client.get_bucket(bucket)
        blob = bucket.blob(name)
        blob.upload_from_filename(path)
    except Exception as e:
        raise UploadError(f"could not upload {name}") from e

    return blob.public_url


def connect_storage(credentials=None):
    """
    Create a google cloud storage client and connect.
    """
    # Attempt to fetch credentials from environment
    credentials = credentials or os.environ.get(GOOGLE_CREDENTIALS, None)

    # Attempt to get credentials from the .secret folder
    credentials = credentials or find_service_account()

    if credentials is None:
        raise UploadError(
            "could not find service account credentials: "
            "set either $GOOGLE_APPLICATION_CREDENTIALS to the path "
            "or store the credentials in the .secret folder"
        )

    # Cannot connect without the storage library.
    if storage is None:
        raise ImportError(
            "the google.cloud.storage module is required, install using pip"
        )

    return storage.Client.from_service_account_json(credentials)


def find_service_account():
    secret = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        "..", "..", ".secret", "*.json"
    ))

    for path in glob.glob(secret):
        with open(path, "r") as f:
            data = json.load(f)
            if "universe_domain" in data and data["universe_domain"] == "googleapis.com":
                return path

    return None



"""
MobileNet benchmark runner
"""

from .exceptions import DatasetsError
from .benchmark import Benchmark, limit_generator
from .models import load_mobilevit, cleanup_mobilevit
from .datasets import load_movies, cleanup_movies, DATASETS


class MobileNet(Benchmark):

    @staticmethod
    def total(**kwargs):
        ## Return the number of movie stills from the manifest
        use_sample = kwargs.pop("use_sample", True)
        name = "movies-sample" if use_sample else "movies"
        if name not in DATASETS:
            raise DatasetsError("movies dataset not found in manifest")
        return DATASETS[name]["instances"]

    @property
    def description(self):
        return (
            "uses the MobileNet v2 model to classify "
            "objects in scenes from movie stills"
        )

    def before(self):
        model, processor = load_mobilevit(model_home=self.model_home)
        self.model = model
        self.processor = processor

    def after(self, cleanup=True):
        if cleanup:
            cleanup_mobilevit(model_home=self.model_home)
            cleanup_movies(data_home=self.data_home, sample=self.use_sample)

    def instances(self, limit=None):
        dataset = load_movies(data_home=self.data_home, sample=self.use_sample)
        return limit_generator(dataset, limit)

    def preprocess(self, instance):
        raise NotImplementedError("MobileNet preprocess not implemented")

    def inference(self, instance):
        raise NotImplementedError("MobileNet inference not implemented")



"""
Benchmark ABC and global benchmark runner.
"""

import time
import dataclasses

from .base import Benchmark
from ..utils import humanize_duration
from ..metrics import Metric, Measurement, dump
from ..exceptions import ConstrueError, BenchmarkError

from tqdm import tqdm
from datetime import datetime, timezone
from typing import Iterable, List, Dict, Optional, Type


DATEFMT = "%Y-%m-%dT%H:%M:%S.%fZ"


class BenchmarkRunner(object):
    """
    Executes one or more benchmarks, configuring them with a top-level config and
    collecting all measurements, merging as necessary and outputing them to a file.
    """

    def __init__(
        self,
        benchmarks: List[Benchmark],
        device: str = None,
        env: str = None,
        n_runs: int = 1,
        limit: int = None,
        data_home: str = None,
        model_home: str = None,
        use_sample: bool = True,
        cleanup: bool = True,
        verbose: bool = True,
    ):
        self.env = env
        self.device = device
        self.n_runs = n_runs
        self.limit = limit
        self.benchmark_kwargs = {
            "data_home": data_home,
            "model_home": model_home,
            "use_sample": use_sample,
            "progress": verbose,
        }
        self.cleanup = cleanup
        self.verbose = verbose
        self.benchmarks = benchmarks

        for b in self.benchmarks:
            if not issubclass(b, Benchmark):
                raise BenchmarkError(f"{b.__name__} is not a Benchmark")

    @property
    def is_complete(self):
        return getattr(self, "run_complete_", False)

    def run(self):
        self.results_ = Results(
            n_runs=self.n_runs,
            limit=self.limit,
            benchmarks=[b.__name__ for b in self.benchmarks],
            started=datetime.now(timezone.utc).strftime(DATEFMT),
            env=self.env,
            device=self.device,
            options=self.benchmark_kwargs,
            errors=[],
        )

        self.run_complete_ = False
        self.measurements_ = []

        started = time.time()

        for cls in self.benchmarks:
            total = self.limit or cls.total(**self.benchmark_kwargs)
            for i in range(self.n_runs):
                self.run_benchmark(i, total, cls)

        self.results_.duration = time.time() - started
        self.results_.measurements = Measurement.merge(self.measurements_)
        self.run_complete_ = True

        if self.verbose:
            print(f"{len(self.benchmarks)} benchmark(s) complete in {humanize_duration(self.results_.duration)}")
            if self.cleanup:
                print("cleaned up data and model caches: all downloaded data removed")

    def run_benchmark(self, idx: int, total: int, Runner: Type):
        # TODO: do we need to pass separate metadata to the kwargs?
        progress = tqdm(total=total, desc=f"Running {Runner.__name__} Benchmark {idx+1}", leave=False)
        benchmark = Runner(**self.benchmark_kwargs)

        try:
            for measurement in self.execute(idx, benchmark, progress):
                self.measurements_.append(measurement)
                self.results_.successes += 1
        except ConstrueError as e:
            self.results_.failures += 1
            self.results_.errors.append(str(e))

    def execute(self, idx: int, benchmark: Benchmark, progress: tqdm) -> Iterable[Measurement]:
        # Setup the benchmark
        benchmark.before()

        ptimes = []  # preproccess times
        itimes = []  # inference times

        try:
            # Time each inference
            # TODO: measure memory usage during inferencing
            for instance in benchmark.instances(limit=self.limit):
                t1 = time.time()
                features = benchmark.preprocess(instance)
                t2 = time.time()
                benchmark.inference(features)
                t3 = time.time()

                ptimes.append(t2 - t1)
                itimes.append(t3 - t2)
                progress.update(1)
        finally:
            # Ensure benchmark is cleaned up despite any errors if this is the last
            # run of the benchmark and cleanup is specified (otherwise leave cache).
            cleanup = self.cleanup and idx == self.n_runs - 1
            benchmark.after(cleanup=cleanup)

        # Create the process times measurement
        yield Measurement(
            per_run=1,
            raw_metrics=ptimes,
            units="s",
            metric=Metric(
                label=benchmark.__class__.__name__,
                sub_label="preprocessing",
                description=benchmark.description,
                device=self.device,
                env=self.env,
            ),
        )

        # Create the inference times measurement
        yield Measurement(
            per_run=1,
            raw_metrics=itimes,
            units="s",
            metric=Metric(
                label=benchmark.__class__.__name__,
                sub_label="inferencing",
                description=benchmark.description,
                device=self.device,
                env=self.env,
            ),
        )

    def save(self, path):
        if not self.is_complete:
            raise BenchmarkError("cannot save benchmarks that haven't been run")

        with open(path, "w") as o:
            dump(self.results_, o)

        if self.verbose:
            print("benchmark results saved to", path)


@dataclasses.dataclass(init=True, repr=False, eq=True)
class Results:
    """
    A result of all runs of a Benchmark including benchmarking information.
    """

    n_runs: int
    benchmarks: List[str]
    started: str
    errors: List[str] = list
    limit: Optional[int] = None
    duration: Optional[float] = None
    env: Optional[str] = None
    device: Optional[str] = None
    options: Optional[Dict] = None
    successes: Optional[int] = 0
    failures: Optional[int] = 0
    measurements: Optional[List[Measurement]] = None



"""
Measurements are the result of a benchmark run.

This module influenced by https://rtnl.link/hMNNI90KBGj
"""

import numpy as np
import dataclasses

from collections import defaultdict
from typing import cast, Optional, Any, Iterable, Dict, List, Tuple


# Measurement will include a warning if the distribution is suspect. All
# runs are expected to have some variation; these parameters set the thresholds.
_IQR_WARN_THRESHOLD = 0.1
_IQR_GROSS_WARN_THRESHOLD = 0.25


@dataclasses.dataclass(init=True, repr=False, eq=True, frozen=True)
class Metric:
    """
    Container information used to define a benchmark measurement.

    This class is similar to a pytorch TaskSpec.
    """

    label: Optional[str] = None
    sub_label: Optional[str] = None
    description: Optional[str] = None
    device: Optional[str] = None
    env: Optional[str] = None

    @property
    def title(self) -> str:
        """
        Best effort attempt at a string label for the metric.
        """
        if self.label is not None:
            return self.label + (f": {self.sub_label}" if self.sub_label else "")
        elif self.env is not None:
            return f"Metric for {self.env}" + (f" on {self.device}" if self.device else "")  # noqa
        return "Metric"

    def summarize(self) -> str:
        """
        Builds a summary string for printing the metric.
        """
        parts = [
            self.title,
            self.description or ""
        ]
        return "\n".join([f"{i}\n" if "\n" in i else i for i in parts if i])


_TASKSPEC_FIELDS = tuple(i.name for i in dataclasses.fields(Metric))


@dataclasses.dataclass(init=True, repr=False)
class Measurement:
    """
    The result of a benchmark measurement.

    This class stores one or more measurements of a given statement. It is similar to
    the pytorch measurement and provides convienence methods and serialization.
    """

    metric: Metric
    raw_metrics: List[float]
    per_run: int = 1
    units: Optional[str] = None
    metadata: Optional[Dict[Any, Any]] = None

    def __post_init__(self) -> None:
        self._sorted_metrics: Tuple[float, ...] = ()
        self._warnings: Tuple[str, ...] = ()
        self._median: float = -1.0
        self._mean: float = -1.0
        self._p25: float = -1.0
        self._p75: float = -1.0

    def __getattr__(self, name: str) -> Any:
        # Forward Metric fields for convenience.
        if name in _TASKSPEC_FIELDS:
            return getattr(self.task_spec, name)
        return super().__getattribute__(name)

    def _compute_stats(self) -> None:
        """
        Comptues the internal stats for the measurements if not already computed.
        """
        if self.raw_metrics and not self._sorted_metrics:
            self._sorted_metrics = tuple(sorted(self.metrics))
            _metrics = np.array(self._sorted_metrics, dtype=np.float64)
            self._median = np.quantile(_metrics, 0.5).item()
            self._mean = _metrics.mean()
            self._p25 = np.quantile(_metrics, 0.25).item()
            self._p75 = np.quantile(_metrics, 0.75).item()

            if not self.meets_confidence(_IQR_GROSS_WARN_THRESHOLD):
                self.__add_warning("This suggests significant environmental influence.")
            elif not self.meets_confidence(_IQR_WARN_THRESHOLD):
                self.__add_warning("This could indicate system fluctuation.")

    def __add_warning(self, msg: str) -> None:
        riqr = self.iqr / self.median * 100
        self._warnings += (
            f"  WARNING: Interquartile range is {riqr:.1f}% "
            f"of the median measurement.\n           {msg}",
        )

    @property
    def metrics(self) -> List[float]:
        return [m / self.per_run for m in self.raw_metrics]

    @property
    def median(self) -> float:
        self._compute_stats()
        return self._median

    @property
    def mean(self) -> float:
        self._compute_stats()
        return self._mean

    @property
    def iqr(self) -> float:
        self._compute_stats()
        return self._p75 - self._p25

    @property
    def has_warnings(self) -> bool:
        self._compute_stats()
        return bool(self._warnings)

    @property
    def title(self) -> str:
        return self.metric.title

    @property
    def env(self) -> str:
        return "Unspecified env" if self.metric.env is None else cast(str, self.metric.env)  # noqa

    @property
    def row_name(self) -> str:
        return self.sub_label or "[Unknown]"

    def meets_confidence(self, threshold: float = _IQR_WARN_THRESHOLD) -> bool:
        return self.iqr / self.median < threshold

    def to_array(self):
        return np.array(self.metrics, dtype=np.float64)

    @staticmethod
    def merge(measurements: Iterable["Measurement"]) -> List["Measurement"]:
        """
        Merge measurement replicas into a single measurement.

        This method will extrapolate per_run=1 and will not transfer metadata.
        """
        groups = defaultdict(list)
        for m in measurements:
            groups[m.metric].append(m)

        def merge_group(metric: Metric, group: List["Measurement"]) -> "Measurement":
            metrics: List[float] = []
            for m in group:
                metrics.extend(m.metrics)

            return Measurement(
                per_run=1,
                raw_metrics=metrics,
                metric=metric,
                metadata=None
            )

        return [merge_group(t, g) for t, g in groups.items()]


def select_duration_unit(t: float) -> Tuple[str, float]:
    """
    Determine how to scale a duration to format for human readability.
    """
    unit = {-3: "ns", -2: "us", -1: "ms"}.get(int(np.log10(np.array(t)).item() // 3), "s")
    scale = {"ns": 1e-9, "us": 1e-6, "ms": 1e-3, "s": 1}[unit]
    return unit, scale


def humanize_duration(u: str) -> str:
    return {
        "ns": "nanosecond",
        "us": "microsecond",
        "ms": "millisecond",
        "s": "second",
    }[u]



"""
Handles serialization and deserialization of metrics.
"""

import json
import dataclasses

from functools import partial

from .metrics import Metric, Measurement


class MetricsJSONEncoder(json.JSONEncoder):

    def default(self, o):
        """
        Encode first looks to see if there is a dump method and uses that, otherwise
        it attempts to serialize a dataclass; and the last step is perform the default
        JSON serialization of primitive types.
        """

        if hasattr(o, "dump"):
            data = o.dump()
            data["type"] = o.__class__.__name__
            return data

        if dataclasses.is_dataclass(o):
            data = dataclasses.asdict(o)
            data["type"] = o.__class__.__name__
            return data

        return super(MetricsJSONEncoder, self).default(o)


class MetricsJSONDecoder(json.JSONDecoder):

    classmap = {
        Metric.__name__: Metric,
        Measurement.__name__: Measurement,
    }

    def __init__(self, *args, **kwargs):
        if not kwargs.get("object_hook", None) is None:
            kwargs["object_hook"] = self.object_hook
        super(MetricsJSONDecoder, self).__init__(*args, **kwargs)

    def object_hook(self, data):
        if "type" in data and data["type"] in self.classmap:
            cls = self.classmap[data.pop("type")]
            if hasattr(cls, "load"):
                return cls.load(data)
            return cls(**data)

        return data


# JSON Serialization
dump = partial(json.dump, cls=MetricsJSONEncoder)
dumps = partial(json.dumps, cls=MetricsJSONEncoder)
load = partial(json.load, cls=MetricsJSONDecoder)
loads = partial(json.loads, cls=MetricsJSONDecoder)



"""
Helper utility for uploading models to GCP
"""

import os

from .path import FIXTURES
from .source import SOURCE_MODELS
from ..version import get_version
from ..utils import resolve_exclude
from ..cloud.gcp import connect_storage, upload
from ..cloud.manifest import make_fixture_path, MODELS


def upload_models(fixtures=FIXTURES, exclude=None, include=None, credentials=None):
    """
    Upload all models from the model fixtures directory.
    """
    exclude = resolve_exclude(exclude, include, SOURCE_MODELS)

    version = get_version(short=True)
    client = connect_storage(credentials)

    for name in SOURCE_MODELS:
        if name in exclude:
            continue

        name += ".zip"
        dst = make_fixture_path(name, MODELS, version=version)
        src = os.path.join(fixtures, name)

        url = upload(dst, src, client)
        print(f"uploaded {url}")



"""
Manifest handlers for datasets
"""

import os
import glob
import json
import shutil
import tempfile

from collections import defaultdict

from .path import FIXTURES, MANIFEST
from ..cloud.manifest import DATASETS
from ..cloud.manifest import load_manifest as _load_manifest
from ..cloud.manifest import generate_manifest as _generate_manifest


def load_manifest(path=MANIFEST):
    return _load_manifest(path)


def generate_manifest(fixtures=FIXTURES, out=MANIFEST):
    out = out or MANIFEST
    return _generate_manifest(fixtures, out, DATASETS, extra=dataset_extra)


def dataset_extra(path, name, **kwargs):
    """
    Count the number of instances in each class in the dataset
    """
    with tempfile.TemporaryDirectory() as data_home:
        basename = os.path.basename(path)
        archive = os.path.join(data_home, basename)
        datadir = os.path.join(data_home, name)

        # Copy and extract the archive
        shutil.copy(path, archive)
        shutil.unpack_archive(archive, datadir)

        extra = {"instances": 0, "classes": defaultdict(int)}
        name = name.removesuffix("-sample")

        # Count the number of instances
        {
            "dialects": _count_files,
            "lowlight": _count_files,
            "reddit": _count_jsonl,
            "movies": _count_files,
            "essays": _count_jsonl,
            "aegis": _count_jsonl,
            "nsfw": _count_files,
        }[name](name, datadir, extra)

        return extra


def _count_files(name, datadir, extra):
    patterns = {
        "lowlight": "lowlight/**/*.png",
        "nsfw": "nsfw/**/*.jpg",
    }

    if name in patterns:
        pattern = patterns[name]
    else:
        pattern = "**/*"

    for path in glob.glob(os.path.join(datadir, pattern)):
        if os.path.isdir(path):
            continue

        label = os.path.basename(os.path.dirname(path))
        extra["instances"] += 1
        extra["classes"][label] += 1


def _count_jsonl(name, datadir, extra):
    for path in glob.glob(os.path.join(datadir, "*.jsonl")):
        with open(path, "r") as f:
            for line in f:
                try:
                    json.loads(line)
                except json.JSONDecodeError:
                    continue

                extra["instances"] += 1

    del extra["classes"]



"""
Utilities for construe
"""

import os

from datetime import timedelta
from typing import Iterable, Set, Union


def resolve_exclude(
    exclude: Iterable[str] = None,
    include: Iterable[str] = None,
    all: Iterable[str] = None,
) -> Set[str]:
    """
    Given an exclusion list and an inclusion list, merge them to produce a definitive
    exclusion list such that if there are any inclusions, then everything not in the
    inclusion list from all is added to the exclusion list. If an item is both in the
    exclusion list and the inclusion list it is excluded.
    """
    exclude = exclude or []
    exclude = set([item.strip().lower() for item in exclude])

    include = include or []
    include = set([
        item.strip().lower() for item in include
    ])

    if include:
        for item in all:
            if item not in include:
                exclude.add(item)

    return exclude


def humanize_duration(duration: Union[int, float, timedelta]) -> str:
    """
    Represent a duration as a human readable string. If an int or a float are passed
    then it is assumed to be in seconds.
    """
    if not isinstance(duration, timedelta):
        duration = timedelta(seconds=duration)

    days = duration.days
    hours, remainder = divmod(duration.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    if days:
        return f"{days}d {hours}h {minutes}m {seconds}s"
    if hours:
        return f"{hours}h {minutes}m {seconds}s"
    if minutes:
        return f"{minutes}m {seconds}s"
    return f"{seconds}s"


def dirsize(path):
    """
    Return the size utilized by the contents of a directory specified by path in bytes.
    """
    bytes = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            if not os.path.islink(fp):
                bytes += os.path.getsize(fp)
    return bytes



"""
Admin utility for managing datasets.
"""

import click

from construe.version import get_version

from .source import download_source_datasets, SOURCE_DATASETS
from .source import sample_source_datasets
from .manifest import generate_manifest
from .upload import upload_datasets
from .path import FIXTURES


CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
}

EPILOG = "This tool is intended for construe developer use and not construe users."


@click.group(context_settings=CONTEXT_SETTINGS, epilog=EPILOG)
@click.version_option(get_version(), message="%(prog)s v%(version)s")
def main():
    pass


@main.command(epilog=EPILOG)
@click.option(
    "-f",
    "--fixtures",
    type=str,
    default=FIXTURES,
    help="path to fixtures directory to generate manifest from",
)
@click.option(
    "-o",
    "--out",
    type=str,
    default=None,
    help="path to write the manifest to",
)
def manifest(fixtures=FIXTURES, out=None):
    """
    Generate a manifest file from local fixtures.
    """
    generate_manifest(fixtures, out)


@main.command(epilog=EPILOG)
@click.option(
    "-f",
    "--fixtures",
    type=str,
    default=FIXTURES,
    help="path to fixtures directory to download source datasets to",
)
@click.option(
    "-e",
    "--exclude",
    type=click.Choice(SOURCE_DATASETS, case_sensitive=False),
    default=None,
    multiple=True,
    help="specify datasets to exclude from source download",
)
@click.option(
    "-i",
    "--include",
    type=click.Choice(SOURCE_DATASETS, case_sensitive=False),
    default=None,
    multiple=True,
    help="specify datasets to explicitly include in source download",
)
def originals(fixtures=FIXTURES, exclude=None, include=None):
    """
    Download original datasets and store them in fixtures.
    """
    download_source_datasets(out=FIXTURES, exclude=exclude, include=include)


@main.command(epilog=EPILOG)
@click.option(
    "-f",
    "--fixtures",
    type=str,
    default=FIXTURES,
    help="path to fixtures directory to find source datasets in",
)
@click.option(
    "-o",
    "--out",
    type=str,
    default=FIXTURES,
    help="directory to write the sampled data to",
)
@click.option(
    "-s",
    "--size",
    type=float,
    default=0.25,
    help="approximate random sample size",
)
@click.option(
    "-S",
    "--suffix",
    type=str,
    default="-sample",
    help="suffix to append to the dataset name",
)
@click.argument(
    "dataset",
    nargs=-1,
    required=True,
    type=click.Choice(SOURCE_DATASETS, case_sensitive=False),
)
def sample(dataset, fixtures=FIXTURES, out=FIXTURES, size=0.25, suffix="-sample"):
    """
    Create a sample dataset from the original that is smaller.
    """
    sample_source_datasets(dataset, fixtures, out, size, suffix)


@main.command(epilog=EPILOG)
@click.option(
    "-f",
    "--fixtures",
    type=str,
    default=FIXTURES,
    help="path to fixtures directory where source datasets have been downloaded",
)
@click.option(
    "-e",
    "--exclude",
    type=click.Choice(SOURCE_DATASETS, case_sensitive=False),
    default=None,
    multiple=True,
    help="specify datasets to exclude from upload",
)
@click.option(
    "-i",
    "--include",
    type=click.Choice(SOURCE_DATASETS, case_sensitive=False),
    default=None,
    multiple=True,
    help="specify datasets to explicitly include in upload",
)
@click.option(
    "-c",
    "--credentials",
    type=str,
    default=None,
    help="path to service account json credentials for upload",
)
def upload(**kwargs):
    """
    Upload datasets to GCP for user downloads.
    """
    upload_datasets(**kwargs)


if __name__ == "__main__":
    main(
        obj={},
        prog_name="python -m construe.datasets",
    )



"""
This package downloads and instantiates the source models from HuggingFace.
"""

import os
import tensorflow as tf

from .path import FIXTURES
from ..utils import resolve_exclude
from .tflite import TFLiteGenerateModel

from transformers import MobileViTImageProcessor, MobileViTForImageClassification
from transformers import WhisperProcessor, TFWhisperForConditionalGeneration
from transformers import AutoModelForImageClassification, AutoImageProcessor
from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers import AutoModelForSequenceClassification
from transformers import ViTImageProcessor

from gliner import GLiNER
from huggingface_hub import from_pretrained_keras


MOONDREAM_ID = "vikhyatk/moondream2"
MOONDREAM_REVISION = "2024-08-26"
MOONDREAM_DIR = "moondream"
MOONDREAM_SAVED_MODEL_DIR = "hf_moondream_saved"

WHISPER_ID = "openai/whisper-tiny.en"
WHISPER_REVISION = "main"
WHISPER_DIR = "whisper"
WHISPER_SAVED_MODEL_DIR = "tf_whisper_saved"

MOBILENET_ID = "google/mobilenet_v2_1.0_224"
MOBILENET_REVISION = "main"
MOBILENET_DIR = "mobilenet"
MOBILENET_SAVED_MODEL_DIR = "hf_mobilenet_saved"

MOBILEVIT_ID = "apple/mobilevit-xx-small"
MOBILEVIT_REVISION = "main"
MOBILEVIT_DIR = "mobilevit"
MOBILEVIT_SAVED_MODEL_DIR = "hf_mobilevit_saved"

NSFW_ID = "Falconsai/nsfw_image_detection"
NSFW_REVISION = "main"
NSFW_DIR = "nsfw"
NSFW_SAVED_MODEL_DIR = "hf_nsfw_saved"

LOWLIGHT_ID = "keras-io/lowlight-enhance-mirnet"
LOWLIGHT_DIR = "lowlight"
LOWLIGHT_SAVED_MODEL_DIR = "tf_lowlight_saved"

OFFENSIVE_ID = "KoalaAI/OffensiveSpeechDetector"
OFFENSIVE_REVISION = "main"
OFFENSIVE_DIR = "offensive"
OFFENSIVE_SAVED_MODEL_DIR = "hf_offensive_saved"

GLINER_ID = "knowledgator/gliner-bi-small-v1.0"
GLINER_DIR = "gliner"
GLINER_SAVED_MODEL_DIR = "hf_gliner_saved"

SOURCE_MODELS = [
    MOONDREAM_DIR,
    WHISPER_DIR,
    MOBILENET_DIR,
    MOBILEVIT_DIR,
    NSFW_DIR,
    LOWLIGHT_DIR,
    OFFENSIVE_DIR,
    GLINER_DIR,
]


##########################################################################
## Source Models
##########################################################################

def download_source_models(out=FIXTURES, exclude=None, include=None):
    """
    Download all source models to the specified directory, excluding any by name.
    """
    downloaders = {
        MOONDREAM_DIR: download_moondream,
        WHISPER_DIR: download_whisper,
        MOBILENET_DIR: download_mobilenet,
        MOBILEVIT_DIR: download_mobilevit,
        NSFW_DIR: download_nsfw,
        LOWLIGHT_DIR: download_lowlight,
        OFFENSIVE_DIR: download_offensive,
        GLINER_DIR: download_gliner,
    }

    exclude = resolve_exclude(exclude, include, SOURCE_MODELS)

    for name, download in downloaders.items():
        if name in exclude:
            continue
        download(out=out)


def download_moondream(
    out=FIXTURES, model_id=MOONDREAM_ID, revision=MOONDREAM_REVISION
):
    """
    Download the moondream model and save it to fixtures.
    """
    model = AutoModelForCausalLM.from_pretrained(
        model_id, trust_remote_code=True, revision=revision
    )
    tokenizer = AutoTokenizer.from_pretrained(model_id, revision=revision)

    path = os.path.join(out, MOONDREAM_DIR, MOONDREAM_SAVED_MODEL_DIR)

    model.save_pretrained(path)
    tokenizer.save_pretrained(path)

    print(f"moondream downloaded to {path}")
    return path


def download_whisper(out=FIXTURES, model_id=WHISPER_ID, revision=WHISPER_REVISION):
    """
    Download the whisper tiny english model and save it to fixtures.
    """
    processor = WhisperProcessor.from_pretrained(model_id, revision=revision)
    model = TFWhisperForConditionalGeneration.from_pretrained(model_id, revision=revision)
    generate = TFLiteGenerateModel(model)

    path = os.path.join(out, WHISPER_DIR, WHISPER_SAVED_MODEL_DIR)

    tf.saved_model.save(generate, path, signatures={"serving_default": generate.serving})
    processor.save_pretrained(path)

    print(f"whisper downloaded to {path}")
    return path


def download_mobilenet(
    out=FIXTURES, model_id=MOBILENET_ID, revision=MOBILENET_REVISION
):
    """
    Download the MobileNet model and save it to fixtures.
    """
    preprocessor = AutoImageProcessor.from_pretrained(model_id, revision=revision)
    model = AutoModelForImageClassification.from_pretrained(
        model_id, revision=revision, trust_remote_code=True
    )

    path = os.path.join(out, MOBILENET_DIR, MOBILENET_SAVED_MODEL_DIR)

    model.save_pretrained(path)
    preprocessor.save_pretrained(path)

    print(f"mobilenet downloaded to {path}")
    return path


def download_mobilevit(
    out=FIXTURES, model_id=MOBILEVIT_ID, revision=MOBILEVIT_REVISION
):
    """
    Download the MobileNet model and save it to fixtures.
    """
    preprocessor = MobileViTImageProcessor.from_pretrained(model_id, revision=revision)
    model = MobileViTForImageClassification.from_pretrained(model_id, revision=revision)

    path = os.path.join(out, MOBILEVIT_DIR, MOBILEVIT_SAVED_MODEL_DIR)

    model.save_pretrained(path)
    preprocessor.save_pretrained(path)

    print(f"mobilevit downloaded to {path}")
    return path


def download_nsfw(out=FIXTURES, model_id=NSFW_ID, revision=NSFW_REVISION):
    """
    Download the NSFW image classification model and save it to fixtures.
    """
    preprocessor = ViTImageProcessor.from_pretrained(model_id, revision=revision)
    model = AutoModelForImageClassification.from_pretrained(model_id, revision=revision)

    path = os.path.join(out, NSFW_DIR, NSFW_SAVED_MODEL_DIR)

    model.save_pretrained(path)
    preprocessor.save_pretrained(path)

    print(f"nsfw downloaded to {path}")
    return path


def download_lowlight(out=FIXTURES, model_id=LOWLIGHT_ID):
    """
    Download the lowlight Keras model and safe it to fixtures.
    """
    model = from_pretrained_keras(model_id)
    path = os.path.join(out, LOWLIGHT_DIR, LOWLIGHT_SAVED_MODEL_DIR)
    model.save(path)

    print(f"lowlight downloaded to {path}")
    return path


def download_offensive(
    out=FIXTURES, model_id=OFFENSIVE_ID, revision=OFFENSIVE_REVISION
):
    """
    Download the offensive speech detector model and save it to fixtures.
    """
    model = AutoModelForSequenceClassification.from_pretrained(
        model_id, trust_remote_code=True, revision=revision
    )
    tokenizer = AutoTokenizer.from_pretrained(model_id, revision=revision)

    path = os.path.join(out, OFFENSIVE_DIR, OFFENSIVE_SAVED_MODEL_DIR)

    model.save_pretrained(path)
    tokenizer.save_pretrained(path)

    print(f"offensive downloaded to {path}")
    return path


def download_gliner(out=FIXTURES, model_id=GLINER_ID):
    """
    Download the gliner model and save it to fixtures.
    """
    model = GLiNER.from_pretrained(model_id)
    path = os.path.join(out, GLINER_DIR, GLINER_SAVED_MODEL_DIR)
    model.save_pretrained(path)

    print(f"gliner downloaded to {path}")
    return path



"""
GLiNER named entity discovery benchmark runner
"""

from .exceptions import DatasetsError
from .models import load_gliner, cleanup_gliner
from .benchmark import Benchmark, limit_generator
from .datasets import load_essays, cleanup_essays, DATASETS


class GLiNER(Benchmark):

    @staticmethod
    def total(**kwargs):
        # Return the number of essays from the manifest
        use_sample = kwargs.pop("use_sample", True)
        name = "essays-sample" if use_sample else "essays"
        if name not in DATASETS:
            raise DatasetsError("essays dataset not found in manifest")
        return DATASETS[name]["instances"]

    @property
    def description(self):
        return (
            "applies the GLiNER model to identify and classify "
            "named entities in long-form essays"
        )

    def before(self):
        model, processor = load_gliner(model_home=self.model_home)
        self.model = model
        self.processor = processor

    def after(self, cleanup=True):
        if cleanup:
            cleanup_gliner(model_home=self.model_home)
            cleanup_essays(data_home=self.data_home, sample=self.use_sample)

    def instances(self, limit=None):
        dataset = load_essays(data_home=self.data_home, sample=self.use_sample)
        return limit_generator(dataset, limit)

    def preprocess(self, instance):
        raise NotImplementedError("GLiNER preprocess not implemented")

    def inference(self, instance):
        raise NotImplementedError("GLiNER inference not implemented")



import pytest

from datetime import timedelta
from construe.utils import resolve_exclude
from construe.utils import humanize_duration


@pytest.mark.parametrize("exclude,include,expected", [
    (
        None, None, set(),
    ),
    (
        ["red", "orange", "black"], None, {"red", "orange", "black"},
    ),
    (
        None, ["red", "orange", "black"], {"blue", "green", "pink", "purple", "violet"},
    ),
    (
        ["red", "orange"], ["purple", "violet"], {"red", "blue", "orange", "green", "pink", "black"},
    ),
    (
        ["red", "orange"], ["red", "purple", "violet"], {"red", "blue", "orange", "green", "pink", "black"},
    )
])
def test_resolve_exclude(exclude, include, expected):
    """
    Test that exclusions are resolved correctly
    """
    ALL = [
        "red", "blue", "green", "pink", "purple", "orange", "violet", "black",
    ]
    resolved = resolve_exclude(exclude, include, ALL)
    assert resolved == expected


@pytest.mark.parametrize("duration,expected", [
    (0, "0s"),
    (59, "59s"),
    (60, "1m 0s"),
    (61, "1m 1s"),
    (3600, "1h 0m 0s"),
    (3661, "1h 1m 1s"),
    (86400, "1d 0h 0m 0s"),
    (90061, "1d 1h 1m 1s"),
    (timedelta(days=1, hours=1, minutes=1, seconds=1), "1d 1h 1m 1s"),
    (timedelta(seconds=59), "59s"),
    (timedelta(minutes=1), "1m 0s"),
    (timedelta(hours=1), "1h 0m 0s"),
])
def test_humanize_duration(duration, expected):
    """
    Test that durations are humanized correctly
    """
    assert humanize_duration(duration) == expected



"""
Path handling for model downloads
"""

import os
import shutil

from pathlib import Path
from ..cloud.signature import sha256sum
from construe.exceptions import ModelsError


# Fixtures is where model data being prepared is stored
FIXTURES = os.path.join(os.path.dirname(__file__), "fixtures")
MANIFEST = os.path.join(os.path.dirname(__file__), "manifest.json")

# Models dir is the location of downloaded model files
MODELSDIR = Path.home() / ".construe" / "models"

# Names of the models
MOONDREAM = "moondream"
WHISPER = "whisper"
MOBILENET = "mobilenet"
MOBILEVIT = "mobilevit"
NSFW = "nsfw"
LOWLIGHT = "lowlight"
OFFENSIVE = "offensive"
GLINER = "gliner"


def get_model_home(path=None):
    """
    Return the path of the Construe models directory. This folder is used by
    model loaders to avoid downloading model parameters several times.

    By default, this folder is in a config directory in the users home folderso the
    model can be can be easily located. Alternatively it can be set by the
    ``$CONSTRUE_MODELS`` environment variable, or programmatically by giving a folder
    path. Note that the ``'~'`` symbol is expanded to the user home directory, and
    environment variables are also expanded when resolving the path.
    """
    if path is None:
        path = os.environ.get("CONSTRUE_MODELS", MODELSDIR)

    path = os.path.expanduser(path)
    path = os.path.expandvars(path)

    if not os.path.exists(path):
        os.makedirs(path)

    return path


def find_model_path(model, model_home=None, fname=None, ext=None, raises=True):
    """
    Looks up the path to the model specified in the models home directory. The storage
    location of the models can be set with the $CONSTRUE_MODELS environment variable.

    If the model is not found a ``ModelsError`` is raised by default.
    """
    # Resolve the root directory that stores the models
    model_home = get_model_home(model_home)

    # Determine the path to the model
    if fname is None:
        if ext is None:
            path = os.path.join(model_home, model)
        else:
            path = os.path.join(model_home, model, "{}{}".format(model, ext))
    else:
        path = os.path.join(model_home, model, fname)

    if not os.path.exists(path):
        if not raises:
            return None

        raise ModelsError(
            f"could not find model at {path} - does it need to be downloaded?"
        )

    return path


def model_exists(model, model_home=None, fname=None, ext=None):
    """
    Checks to see if the specified model exists in the model home directory.
    """
    path = find_model_path(model, model_home, fname, ext, False)
    if path is not None:
        return os.path.exists(path)
    return False


def model_tflite_exists(model, model_home):
    """
    Checks to see if the model .tflite file exists or not.
    """
    return model_exists(model, model_home=model_home, ext=".tflite")


def model_archive(model, signature, model_home=None, ext=".zip"):
    """
    Checks to see if the model archive file exists and determines if it is the latest
    version by comparing the signature specified with the archive signature.
    """
    model_home = get_model_home(model_home)
    path = os.path.join(model_home, model+ext)

    if os.path.exists(path) and os.path.isfile(path):
        return sha256sum(path) == signature
    return False


def cleanup_model(model, model_home=None, archive=".zip"):
    removed = 0
    model_home = get_model_home(model_home)

    # Paths to remove
    datadir = os.path.join(model_home, model)
    archive = os.path.join(model_home, model+archive)

    if os.path.exists(datadir):
        shutil.rmtree(datadir)
        removed += 1

    if os.path.exists(archive):
        os.remove(archive)
        removed += 1

    return removed



"""
Manifest handlers for models
"""

from ..cloud.manifest import MODELS
from .path import FIXTURES, MANIFEST
from ..cloud.manifest import load_manifest as _load_manifest
from ..cloud.manifest import generate_manifest as _generate_manifest


def load_manifest(path=MANIFEST):
    return _load_manifest(path)


def generate_manifest(fixtures=FIXTURES, out=MANIFEST):
    out = out or MANIFEST
    return _generate_manifest(fixtures, out, MODELS)



"""
Convert source models into tflite format.
"""

import os
import shutil
import tensorflow as tf

from .path import FIXTURES
from ..utils import resolve_exclude

from .source import SOURCE_MODELS
from .source import MOONDREAM_DIR, MOONDREAM_SAVED_MODEL_DIR
from .source import WHISPER_DIR, WHISPER_SAVED_MODEL_DIR
from .source import MOBILENET_DIR, MOBILENET_SAVED_MODEL_DIR
from .source import MOBILEVIT_DIR, MOBILEVIT_SAVED_MODEL_DIR
from .source import NSFW_DIR, NSFW_SAVED_MODEL_DIR
from .source import LOWLIGHT_DIR, LOWLIGHT_SAVED_MODEL_DIR
from .source import OFFENSIVE_DIR, OFFENSIVE_SAVED_MODEL_DIR
from .source import GLINER_DIR, GLINER_SAVED_MODEL_DIR

from transformers import MobileViTImageProcessor, MobileViTForImageClassification
from transformers import AutoModelForImageClassification, AutoImageProcessor
from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers import AutoModelForSequenceClassification
from transformers import ViTImageProcessor
from transformers import WhisperProcessor

from gliner import GLiNER
from huggingface_hub import from_pretrained_keras

MOONDREAM_TFLITE = "moondream.tflite"
WHISPER_TFLITE = "whisper.tflite"
MOBILENET_TFLITE = "mobilenet.tflite"
MOBILEVIT_TFLITE = "mobilevit.tflite"
NSFW_TFLITE = "nsfw.tflite"
LOWLIGHT_TFLITE = "lowlight.tflite"
OFFENSIVE_TFLITE = "offensive.tflite"
GLINER_TFLITE = "gliner.tflite"


def convert_source_models(out=FIXTURES, exclude=None, include=None):
    """
    Convert all source models to tflite format, excluding any by name.
    """
    converter = {
        MOONDREAM_DIR: convert_moondream,
        WHISPER_DIR: convert_whisper,
        MOBILENET_DIR: convert_mobilenet,
        MOBILEVIT_DIR: convert_mobilevit,
        NSFW_DIR: convert_nsfw,
        LOWLIGHT_DIR: convert_lowlight,
        OFFENSIVE_DIR: convert_offensive,
        GLINER_DIR: convert_gliner,
    }

    exclude = resolve_exclude(exclude, include, SOURCE_MODELS)

    for name, convert in converter.items():
        if name in exclude:
            continue
        convert(out=out)


def convert_moondream(out=FIXTURES):
    """
    Convert the moondream model to tflite and save it to fixtures.
    """
    path = os.path.join(out, MOONDREAM_DIR, MOONDREAM_SAVED_MODEL_DIR)
    model = AutoModelForCausalLM.from_pretrained(path, trust_remote_code=True)
    tokenizer = AutoTokenizer.from_pretrained(path)

    assert model is not None
    assert tokenizer is not None

    out = os.path.join(out, MOONDREAM_DIR, MOBILENET_TFLITE)
    print(f"moondream converted to {out}")


def convert_whisper(out=FIXTURES):
    """
    Convert the whisper tiny english model to tflite and save it to fixtures.
    """
    path = os.path.join(out, WHISPER_DIR, WHISPER_SAVED_MODEL_DIR)

    # Convert the model
    converter = tf.lite.TFLiteConverter.from_saved_model(path)
    converter.target_spec.supported_ops = [
        tf.lite.OpsSet.TFLITE_BUILTINS,  # enable TensorFlow Lite ops.
        tf.lite.OpsSet.SELECT_TF_OPS     # enable TensorFlow ops.
    ]
    converter.optimizations = [tf.lite.Optimize.DEFAULT]
    tflite = converter.convert()

    converted = os.path.join(out, WHISPER_DIR, WHISPER_DIR)
    os.mkdir(converted)

    tfpath = os.path.join(converted, WHISPER_TFLITE)
    with open(tfpath, 'wb') as f:
        f.write(tflite)

    # Load and save the whisper preprocessor
    processor = WhisperProcessor.from_pretrained(path)
    processor.save_pretrained(converted)

    # Zip the model for packaging
    zipname = os.path.join(out, WHISPER_DIR)
    shutil.make_archive(zipname, "zip", converted)
    print(f"whisper converted to {zipname+'.zip'}")


def convert_mobilenet(out=FIXTURES):
    """
    Convert the mobilenet model to tflite and save it to fixtures.
    """
    path = os.path.join(out, MOBILENET_DIR, MOBILENET_SAVED_MODEL_DIR)
    preprocessor = AutoImageProcessor.from_pretrained(path)
    model = AutoModelForImageClassification.from_pretrained(
        path, trust_remote_code=True
    )

    assert preprocessor is not None
    assert model is not None

    out = os.path.join(out, MOBILENET_DIR, MOBILENET_TFLITE)
    print(f"mobilenet converted to {out}")


def convert_mobilevit(out=FIXTURES):
    """
    Convert the mobilevit model to tflite and save it to fixtures.
    """
    path = os.path.join(out, MOBILEVIT_DIR, MOBILEVIT_SAVED_MODEL_DIR)
    preprocessor = MobileViTImageProcessor.from_pretrained(path)
    model = MobileViTForImageClassification.from_pretrained(path)

    assert preprocessor is not None
    assert model is not None

    out = os.path.join(out, MOBILEVIT_DIR, MOBILEVIT_TFLITE)
    print(f"mobilevit converted to {out}")


def convert_nsfw(out=FIXTURES):
    """
    Convert the mobilevit model to tflite and save it to fixtures.
    """
    path = os.path.join(out, NSFW_DIR, NSFW_SAVED_MODEL_DIR)
    preprocessor = ViTImageProcessor.from_pretrained(path)
    model = AutoModelForImageClassification.from_pretrained(path)

    assert preprocessor is not None
    assert model is not None

    out = os.path.join(out, NSFW_DIR, NSFW_TFLITE)
    print(f"nsfw converted to {out}")


def convert_lowlight(out=FIXTURES):
    """
    Convert the mobilevit model to tflite and save it to fixtures.
    """
    path = os.path.join(out, LOWLIGHT_DIR, LOWLIGHT_SAVED_MODEL_DIR)
    model = from_pretrained_keras(path)

    # Convert the model
    converter = tf.lite.TFLiteConverter.from_keras_model(model)
    converter.target_spec.supported_ops = [
        tf.lite.OpsSet.TFLITE_BUILTINS,  # enable TensorFlow Lite ops.
        tf.lite.OpsSet.SELECT_TF_OPS     # enable TensorFlow ops.
    ]
    converter.optimizations = [tf.lite.Optimize.DEFAULT]
    tflite = converter.convert()

    converted = os.path.join(out, LOWLIGHT_DIR, LOWLIGHT_DIR)
    os.mkdir(converted)

    tfpath = os.path.join(converted, LOWLIGHT_TFLITE)
    with open(tfpath, 'wb') as f:
        f.write(tflite)

    # Zip the model for packaging
    zipname = os.path.join(out, LOWLIGHT_DIR)
    shutil.make_archive(zipname, "zip", converted)
    print(f"lowlight converted to {zipname+'.zip'}")


def convert_offensive(out=FIXTURES):
    """
    Convert the mobilevit model to tflite and save it to fixtures.
    """
    path = os.path.join(out, OFFENSIVE_DIR, OFFENSIVE_SAVED_MODEL_DIR)
    model = AutoModelForSequenceClassification.from_pretrained(
        path, trust_remote_code=True
    )
    tokenizer = AutoTokenizer.from_pretrained(path)

    assert model is not None
    assert tokenizer is not None

    out = os.path.join(out, OFFENSIVE_DIR, OFFENSIVE_TFLITE)
    print(f"offensive converted to {out}")


def convert_gliner(out=FIXTURES):
    """
    Convert the mobilevit model to tflite and save it to fixtures.
    """
    path = os.path.join(out, GLINER_DIR, GLINER_SAVED_MODEL_DIR)
    model = GLiNER.from_pretrained(path)

    assert model is not None

    out = os.path.join(out, GLINER_DIR, GLINER_TFLITE)
    print(f"gliner converted to {out}")



"""
Benchmarks basic dot product torch operators.

See: https://pytorch.org/tutorials/recipes/recipes/benchmark.html
"""

import tqdm
import torch
import pickle
import torch.utils.benchmark as benchmark

from itertools import product
from torch.utils.benchmark import Fuzzer, FuzzedParameter, FuzzedTensor


def batched_dot_mul_sum(a, b):
    """
    Computes batched dot by multiplying and summing
    """
    return a.mul(b).sum(-1)


def batched_dot_bmm(a, b):
    """
    Computes batched dot by reducing to bmm
    """
    a = a.reshape(-1, 1, a.shape[-1])
    b = b.reshape(-1, b.shape[-1], 1)
    return torch.bmm(a, b).flatten(-3)


class BasicBenchmark(object):

    def __init__(self, env=None, saveto=None, num_threads=None, fuzz=False, seed=None):
        if num_threads is None:
            num_threads = torch.get_num_threads()

        self.env = env
        self.saveto = saveto
        self.num_threads = num_threads
        self.fuzz = fuzz
        self.seed = seed

    def run(self):
        results = []
        dataset = self.fuzzer().take(10) if self.fuzz else list(self.static())

        kwargs = {
            "label": "Batched Dot",
            "num_threads": self.num_threads,
            "env": self.env,
        }

        for tensors, tensor_params, params in tqdm.tqdm(dataset, leave=False):
            sub_label = f"{params['k0']:<6} x {params['k1']:<4} {'' if tensor_params['x']['is_contiguous'] else '(discontiguous)'}"  # noqa
            results.append(
                benchmark.Timer(
                    stmt="batched_dot_mul_sum(x, x)",
                    setup="from construe.basic import batched_dot_mul_sum",
                    globals=tensors,
                    sub_label=sub_label,
                    description="mul/sum",
                    **kwargs
                ).blocked_autorange(min_run_time=1)
            )
            results.append(
                benchmark.Timer(
                    stmt="batched_dot_bmm(x, x)",
                    setup="from construe.basic import batched_dot_bmm",
                    globals=tensors,
                    sub_label=sub_label,
                    description="bmm",
                    **kwargs
                ).blocked_autorange(min_run_time=1)
            )

        if self.saveto is not None:
            with open(self.saveto, "wb") as f:
                pickle.dump(results, f)

        compare = benchmark.Compare(results)
        compare.print()

    def fuzzer(self):
        """
        Generates random tensors with 128 to 10000000 elements and sizes k0 and k1
        chosen from a loguniform distribution in [1, 10000], 40% of which will be
        discontiguous on average.
        """
        return Fuzzer(
            parameters=[
                FuzzedParameter(
                    "k0", minval=1, maxval=10000, distribution="loguniform"
                ),
                FuzzedParameter(
                    "k1", minval=1, maxval=10000, distribution="loguniform"
                ),
            ],
            tensors=[
                FuzzedTensor(
                    "x",
                    size=("k0", "k1"),
                    min_elements=128,
                    max_elements=10000000,
                    probability_contiguous=0.6,
                )
            ],
            seed=self.seed,
        )

    def static(self):
        sizes = [16, 64, 1024, 16384]
        for k0, k1 in product(sizes, sizes):
            params = {
                "k0": k0,
                "k1": k1,
            }

            tensors = {
                "x": torch.ones((k0, k1))
            }

            tensor_params = {
                "x": {
                    "is_contiguous": True
                }
            }

            yield tensors, tensor_params, params



"""
Whisper benchmark runner
"""

import soundfile as sf

from .exceptions import DatasetsError
from .exceptions import InferenceError
from .benchmark import Benchmark, limit_generator
from .models import load_whisper, cleanup_whisper
from .datasets import load_dialects, cleanup_dialects, DATASETS


class Whisper(Benchmark):

    @staticmethod
    def total(**kwargs):
        # Return the number of dialects from the manifest
        use_sample = kwargs.pop("use_sample", True)
        name = "dialects-sample" if use_sample else "dialects"
        if name not in DATASETS:
            raise DatasetsError("dialects dataset not found in manifest")
        return DATASETS[name]["instances"]

    @property
    def description(self):
        return (
            "utilizes the whisper-tiny english model to "
            "transcribe audio from various UK dialects"
        )

    def before(self):
        model, processor = load_whisper(model_home=self.model_home)
        self.model = model
        self.processor = processor
        self.generate = self.model.get_signature_runner()

    def after(self, cleanup=True):
        if cleanup:
            cleanup_whisper(model_home=self.model_home)
            cleanup_dialects(data_home=self.data_home, sample=self.use_sample)

    def instances(self, limit=None):
        dataset = load_dialects(data_home=self.data_home, sample=self.use_sample)
        return limit_generator(dataset, limit)

    def preprocess(self, instance):
        # Instance is a path to a a sound file on disk.
        try:
            audio, samplerate = sf.read(instance)
        except Exception as e:
            raise InferenceError("could not extract audio from file") from e

        audio = audio.astype('float64')
        audio = audio / 32767.0

        return self.processor(audio, sampling_rate=samplerate, return_tensors="tf")

    def inference(self, instance):
        audio = instance.input_features
        sequences = self.generate(input_features=audio)["sequences"]
        return self.processor.batch_decode(sequences, skip_special_tokens=True)



"""
Helper utility for uploading datasets to GCP
"""

import os

from .path import FIXTURES
from .source import SOURCE_DATASETS
from ..version import get_version
from ..utils import resolve_exclude
from ..cloud.gcp import connect_storage, upload
from ..cloud.manifest import make_fixture_path, DATASETS


def upload_datasets(fixtures=FIXTURES, exclude=None, include=None, credentials=None):
    """
    Upload all datasets from the model fixtures directory.
    """
    exclude = resolve_exclude(exclude, include, SOURCE_DATASETS)

    version = get_version(short=True)
    client = connect_storage(credentials)

    for name in SOURCE_DATASETS:
        if name in exclude:
            continue

        for fname in (name + ".zip", name + "-sample.zip"):
            dst = make_fixture_path(fname, DATASETS, version=version)
            src = os.path.join(fixtures, fname)
            url = upload(dst, src, client)
            print(f"uploaded {url}")



"""
NSFW Image Classification benchmark runner
"""

from .datasets import DATASETS
from .exceptions import DatasetsError
from .benchmark import Benchmark, limit_generator
from .models import load_nsfw as load_nsfw_model
from .models import cleanup_nsfw as cleanup_nsfw_model
from .datasets import load_nsfw as load_nsfw_dataset
from .datasets import cleanup_nsfw as cleanup_nsfw_dataset


class NSFW(Benchmark):

    @staticmethod
    def total(**kwargs):
        # Return the number of nsfw images from the manifest
        use_sample = kwargs.pop("use_sample", True)
        name = "nsfw-sample" if use_sample else "nsfw"
        if name not in DATASETS:
            raise DatasetsError("nsfw dataset not found in manifest")
        return DATASETS[name]["instances"]

    @property
    def description(self):
        return (
            "uses a fine-tuned model to classify images as "
            "safe or not safe for work (nsfw)"
        )

    def before(self):
        model, processor = load_nsfw_model(model_home=self.model_home)
        self.model = model
        self.processor = processor

    def after(self, cleanup=True):
        if cleanup:
            cleanup_nsfw_model(model_home=self.model_home)
            cleanup_nsfw_dataset(data_home=self.data_home, sample=self.use_sample)

    def instances(self, limit=None):
        dataset = load_nsfw_dataset(data_home=self.data_home, sample=self.use_sample)
        return limit_generator(dataset, limit)

    def preprocess(self, instance):
        raise NotImplementedError("NSFW preprocess not implemented")

    def inference(self, instance):
        raise NotImplementedError("NSFW inference not implemented")



"""
Manages datasets used for inferencing
"""

from .loaders import * # noqa
from .download import download_data, load_manifest
from .path import get_data_home, cleanup_dataset

try:
    DATASETS = load_manifest()
except Exception:
    DATASETS = None



"""
Top-level exceptions for handling construe errors.
"""

import re

from click import ClickException


class ConstrueError(ClickException):
    pass


class DownloadError(ConstrueError):
    pass


class UploadError(ConstrueError):
    pass


class DatasetsError(ConstrueError):
    pass


class ModelsError(ConstrueError):
    pass


class InferenceError(ConstrueError):
    pass


class BenchmarkError(ConstrueError):
    pass


class DeviceError(ConstrueError):

    def __init__(self, e):
        """
        Expects a runtime error to parse from PyTorch
        """
        tre = re.compile(r"^Expected one of ([\w,\s]+) device type at start of device string: ([\w\d:]+)$", re.I) # noqa
        match = tre.match(str(e))
        if match:
            unknown = match.groups()[1]
            devices = match.groups()[0]
            super(DeviceError, self).__init__((
                f"Cannot set pytorch device (unknown device \"{unknown}\")\n"
                f"Device must be one of the following:\n{devices}"
            ))
        else:
            super(DeviceError, self).__init__(str(e))



"""
This module handles downloading data from original data sources and converting them
into the source format expected by the construe library.
"""

import os
import json
import random
import shutil
import zipfile
import soundfile as sf

from .path import FIXTURES
from ..cloud.download import CHUNK

from datasets import Audio
from functools import partial
from datasets import load_dataset
from urllib.request import urlopen
from click.exceptions import UsageError


AEGIS = "aegis"
AEGIS_BASENAME = "aegis.zip"
AEGIS_ZIPNAME = "aegis.jsonl"
AEGIS_DATASET_URL = "nvidia/Aegis-AI-Content-Safety-Dataset-1.0"

LOWLIGHT = "lowlight"
LOWLIGHT_BASENAME = "lowlight.zip"
LOWLIGHT_DATASET_URL = (
    "https://drive.google.com/uc?id=1DdGIJ4PZPlF2ikl8mNM9V-PdVxVLbQi6"
)

DIALECTS = "dialects"
DIALECTS_BASENAME = "dialects.zip"
DIALECTS_DATASET_URL = "ylacombe/english_dialects"

REDDIT = "reddit"
REDDIT_BASENAME = "reddit.zip"
REDDIT_ZIPNAME = "reddit.jsonl"
REDDIT_DATASETS_URL = "ummagumm-a/reddit_posts_comments"

ESSAYS = "essays"
ESSAYS_BASENAME = "essays.zip"
ESSAYS_ZIPNAME = "essays.jsonl"
ESSAYS_DATASETS_URL = "knarasi1/student_and_llm_essays"

NSFW = "nsfw"
NSFW_BASENAME = "nsfw.zip"
NSFW_DATASETS_URL = "zanderlewis/nsfw_detection_large"

MOVIES = "movies"
MOVIES_BASENAME = "movies.zip"
MOVIES_DATASETS_URL = "unography/movie-scenes"

SOURCE_DATASETS = [
    AEGIS, LOWLIGHT, DIALECTS, REDDIT, ESSAYS, NSFW, MOVIES
]


##########################################################################
## Downloaders
##########################################################################

def download_source_datasets(out=FIXTURES, exclude=None):
    """
    Download all source datasets to the specified out directory,
    excluding any by name in the downloaders dictionary.
    """
    downloaders = {
        AEGIS: download_aegis,
        LOWLIGHT: download_lowlight,
        DIALECTS: download_dialects,
        REDDIT: download_reddit,
        ESSAYS: download_essays,
        NSFW: download_nsfw,
        MOVIES: download_movies,
    }

    exclude = exclude or []
    exclude = set([
        item.strip().lower() for item in exclude
    ])

    # Always exclude lowlight for now
    exclude.add("lowlight")

    for name, download in downloaders.items():
        if name in exclude:
            continue
        download(out=out)


def download_aegis(out=FIXTURES):
    """
    Download the AEGIS AI safety dataset from hugging faced and save as
    a zip compressed json lines file.
    """
    path = os.path.join(out, AEGIS_BASENAME)
    ds = load_dataset(AEGIS_DATASET_URL, split=None)
    with zipfile.ZipFile(path, "x", compression=zipfile.ZIP_DEFLATED) as z:
        with z.open(AEGIS_ZIPNAME, "w") as d:
            for split in ds.values():
                for row in split:
                    d.write(json.dumps(row).encode("utf-8"))
                    d.write("\n".encode("utf-8"))


def download_lowlight(url=LOWLIGHT_DATASET_URL, out=FIXTURES):
    """
    Download the lowlight dataset from Google Drive.

    NOTE: this dataset needs to be downloaded manually from a browser
    because a javascript button click is necessary to start the download.
    """
    path = os.path.join(out, LOWLIGHT_BASENAME)

    response = urlopen(url)
    with open(path, "wb") as f:
        while True:
            chunk = response.read(CHUNK)
            if not chunk:
                break
            f.write(chunk)

    # TODO: unzip the download and create only one folder with all
    # the high and low photos without separation.


def download_dialects(out=FIXTURES):
    configs = [
        "irish_male",
        "midlands_female",
        "midlands_male",
        "northern_female",
        "northern_male",
        "scottish_female",
        "scottish_male",
        "southern_female",
        "southern_male",
        "welsh_female",
        "welsh_male",
    ]

    dirname, _ = os.path.splitext(DIALECTS_BASENAME)
    dir = os.path.join(out, dirname)
    if not os.path.exists(dir):
        os.makedirs(dir)

    for config in configs:
        ds = load_dataset(DIALECTS_DATASET_URL, config)
        cdir = os.path.join(dir, config)
        if not os.path.exists(cdir):
            os.makedirs(cdir)

        for split in ds.values():
            split = split.cast_column("audio", Audio(sampling_rate=16000))
            for row in split:
                path = os.path.join(cdir, row["audio"]["path"])
                audio = row["audio"]["array"] * 32767
                audio = audio.astype('int16')
                sf.write(path, audio, 16000)

    shutil.make_archive(dir, "zip", dir)


def download_reddit(out=FIXTURES):
    path = os.path.join(out, REDDIT_BASENAME)
    ds = load_dataset(REDDIT_DATASETS_URL, split=None)
    with zipfile.ZipFile(path, "x", compression=zipfile.ZIP_DEFLATED) as z:
        with z.open(REDDIT_ZIPNAME, "w") as d:
            for split in ds.values():
                for row in split:
                    d.write(json.dumps(row).encode("utf-8"))
                    d.write("\n".encode("utf-8"))


def download_essays(out=FIXTURES):
    path = os.path.join(out, ESSAYS_BASENAME)
    ds = load_dataset(ESSAYS_DATASETS_URL, split=None)
    with zipfile.ZipFile(path, "x", compression=zipfile.ZIP_DEFLATED) as z:
        with z.open(ESSAYS_ZIPNAME, "w") as d:
            for split in ds.values():
                for row in split:
                    d.write(json.dumps(row).encode("utf-8"))
                    d.write("\n".encode("utf-8"))


def download_nsfw(out=FIXTURES):
    img = 0
    path = os.path.join(out, NSFW_BASENAME)
    fname, _ = os.path.splitext(NSFW_BASENAME)
    ds = load_dataset(NSFW_DATASETS_URL, split=None)

    with zipfile.ZipFile(path, "x", compression=zipfile.ZIP_DEFLATED) as z:
        for split in ds.values():
            for row in split:
                img += 1
                folder = "nsfw" if row["label"] == 0 else "safe"
                imgpath = os.path.join(fname, folder, f"img{img:0>3}.jpg")
                with z.open(imgpath, "w") as d:
                    image = row["image"]
                    if image.mode != "RGB":
                        image = image.convert("RGB")
                    image.save(d, format="jpeg")


def download_movies(out=FIXTURES):
    img = 0
    path = os.path.join(out, MOVIES_BASENAME)
    fname, _ = os.path.splitext(MOVIES_BASENAME)
    ds = load_dataset(MOVIES_DATASETS_URL, split=None)

    with zipfile.ZipFile(path, "x", compression=zipfile.ZIP_DEFLATED) as z:
        for split in ds.values():
            for row in split:
                img += 1
                imgpath = os.path.join(fname, f"img{img:0>6}.jpg")
                with z.open(imgpath, "w") as d:
                    image = row["image"]
                    if image.mode != "RGB":
                        image = image.convert("RGB")
                    image.save(d, format="jpeg")


##########################################################################
## Sampling
##########################################################################


def sample_source_datasets(
    datasets, fixtures=FIXTURES, out=FIXTURES, size=0.25, suffix="-sample"
):
    """
    Sample the specified datasets creating a smaller dataset for benchmarking.
    """
    samplers = {
        AEGIS: sample_aegis,
        LOWLIGHT: sample_lowlight,
        DIALECTS: sample_dialects,
        REDDIT: sample_reddit,
        ESSAYS: sample_essays,
        NSFW: sample_nsfw,
        MOVIES: sample_movies,
    }

    # Check that the samplers are available
    for dataset in datasets:
        dataset = dataset.strip().lower()
        if dataset not in samplers:
            raise UsageError(f"unknown dataset '{dataset}': cannot run sampler")

    # Execute samplers
    for dataset in datasets:
        dataset = dataset.strip().lower()
        samplers[dataset](fixtures, out, size, suffix)


def _sample_files(name, fixtures=FIXTURES, out=FIXTURES, size=0.25, suffix="-sample"):
    files_read, files_written = 0, 0
    inpath, outpath = _sample_paths(name, fixtures, out, suffix)

    with zipfile.ZipFile(inpath, "r") as zi:
        with zipfile.ZipFile(outpath, "x", compression=zipfile.ZIP_DEFLATED) as zo:
            for fp in zi.infolist():
                if fp.is_dir() or fp.filename.startswith("."):
                    continue

                with zi.open(fp.filename, "r") as f:
                    files_read += 1
                    if random.random() <= size:
                        with zo.open(fp.filename, "w") as o:
                            o.write(f.read())
                            files_written += 1

    print(f"sample {outpath} wrote {files_written} out of {files_read}")


def _sample_jsonl(name, fixtures=FIXTURES, out=FIXTURES, size=0.25, suffix="-sample"):
    fname, _ = os.path.splitext(name)
    fname += ".jsonl"

    lines_read, lines_written = 0, 0

    inpath, outpath = _sample_paths(name, fixtures, out, suffix)
    with zipfile.ZipFile(inpath, "r") as zi:
        with zipfile.ZipFile(outpath, "x", compression=zipfile.ZIP_DEFLATED) as zo:
            with zi.open(fname, "r") as f:
                with zo.open(fname, "w") as o:
                    for line in f:
                        lines_read += 1
                        if random.random() <= size:
                            o.write(line)
                            lines_written += 1

    print(f"sample {outpath} wrote {lines_written} out of {lines_read}")


def _sample_paths(name, fixtures, out, suffix):
    inpath = os.path.join(fixtures, name)
    basename, ext = os.path.splitext(name)
    outpath = os.path.join(out, f"{basename}{suffix}{ext}")

    if os.path.exists(outpath):
        os.remove(outpath)

    return inpath, outpath


sample_aegis = partial(_sample_jsonl, AEGIS_BASENAME)
sample_lowlight = partial(_sample_files, LOWLIGHT_BASENAME)
sample_dialects = partial(_sample_files, DIALECTS_BASENAME)
sample_reddit = partial(_sample_jsonl, REDDIT_BASENAME)
sample_essays = partial(_sample_jsonl, ESSAYS_BASENAME)
sample_nsfw = partial(_sample_files, NSFW_BASENAME)
sample_movies = partial(_sample_files, MOVIES_BASENAME)



"""
Managers for loading datasets
"""

import os
import glob
import json
import shutil

from functools import partial

from .manifest import load_manifest
from .download import download_data
from ..exceptions import DatasetsError
from .path import find_dataset_path, get_data_home
from .path import dataset_archive, cleanup_dataset
from .path import DIALECTS, LOWLIGHT, REDDIT, MOVIES, ESSAYS, AEGIS, NSFW


__all__ = [
    "load_all_datasets", "cleanup_all_datasets",
    "load_dialects", "cleanup_dialects",
    "load_lowlight", "cleanup_lowlight",
    "load_reddit", "cleanup_reddit",
    "load_movies", "cleanup_movies",
    "load_essays", "cleanup_essays",
    "load_aegis", "cleanup_aegis",
    "load_nsfw", "cleanup_nsfw",
]


DATASETS = load_manifest()


def _info(dataset):
    if dataset not in DATASETS:
        raise DatasetsError(f"no dataset named {dataset} exists")
    return DATASETS[dataset]


def _load_prepare(name, sample=True, data_home=None):
    if sample and not name.endswith("-sample"):
        name = name + "-sample"

    info = _info(name)
    if not dataset_archive(name, info["signature"], data_home=data_home):
        # If the dataset does not exist, download and extract it
        kwargs = {
            "data_home": data_home, "replace": True, "extract": True,
            "url": info["url"], "signature": info["signature"],
        }
        download_data(**kwargs)

    return find_dataset_path(name, data_home=data_home, fname=None, ext=None)


def _load_file_dataset(name, sample=True, data_home=None, no_dirs=True, pattern=None):
    # Find the data path
    data_path = _load_prepare(name, sample=sample, data_home=data_home)

    # Glob pattern for discovering files in the dataset
    if pattern is None:
        pattern = os.path.join(data_path, "**", "*")
    else:
        pattern = os.path.join(data_path, pattern)

    for path in glob.glob(pattern):
        if no_dirs and os.path.isdir(path):
            continue

        yield path


def _load_jsonl_dataset(name, sample=True, data_home=None):
    data_path = _load_prepare(name, sample=sample, data_home=data_home)
    for path in glob.glob(os.path.join(data_path, "*.jsonl")):
        with open(path, "r") as f:
            for line in f:
                yield json.loads(line.strip())


def _cleanup_dataset(name, sample=True, data_home=None):
    if sample and not name.endswith("-sample"):
        name = name + "-sample"
    return cleanup_dataset(name, data_home=data_home)


load_dialects = partial(_load_file_dataset, DIALECTS)
cleanup_dialects = partial(_cleanup_dataset, DIALECTS)

load_lowlight = partial(_load_file_dataset, LOWLIGHT, pattern="lowlight/**/*.png")
cleanup_lowlight = partial(_cleanup_dataset, LOWLIGHT)

load_reddit = partial(_load_jsonl_dataset, REDDIT)
cleanup_reddit = partial(_cleanup_dataset, REDDIT)

load_movies = partial(_load_file_dataset, MOVIES)
cleanup_movies = partial(_cleanup_dataset, MOVIES)

load_essays = partial(_load_jsonl_dataset, ESSAYS)
cleanup_essays = partial(_cleanup_dataset, ESSAYS)

load_aegis = partial(_load_jsonl_dataset, AEGIS)
cleanup_aegis = partial(_cleanup_dataset, AEGIS)

load_nsfw = partial(_load_file_dataset, NSFW, pattern="nsfw/**/*.jpg")
cleanup_nsfw = partial(_cleanup_dataset, NSFW)


def load_all_datasets(sample=True, data_home=None):
    """
    Load all available datasets as defined by __all__
    """
    module = globals()
    for name in __all__:
        if not name.startswith("load"):
            continue

        if name == "load_all_datasets":
            continue

        f = module[name]
        for row in f(sample=sample, data_home=data_home):
            yield row


def cleanup_all_datasets(data_home=None):
    """
    Delete everything in the data home directory
    """
    with os.scandir(get_data_home(data_home)) as entries:
        for entry in entries:
            if entry.is_dir() and not entry.is_symlink():
                shutil.rmtree(entry.path)
            else:
                os.remove(entry.path)



"""
Computes the signature of downloaded files for hash verification.
"""

import hashlib


def sha256sum(path, blocksize=65536):
    """
    Computes the SHA256 signature of a file to verify that the file has not
    been modified in transit and that it is the correct version of the data.
    """
    sig = hashlib.sha256()
    with open(path, "rb") as f:
        buf = f.read(blocksize)
        while len(buf) > 0:
            sig.update(buf)
            buf = f.read(blocksize)
    return sig.hexdigest()



"""
Path handling for downloads
"""

import os
import shutil

from pathlib import Path
from ..cloud.signature import sha256sum
from construe.exceptions import DatasetsError


# Fixtures is where data being prepared is stored
FIXTURES = os.path.join(os.path.dirname(__file__), "fixtures")
MANIFEST = os.path.join(os.path.dirname(__file__), "manifest.json")

# Data dir is the location of downloaded datasets
DATADIR = Path.home() / ".construe" / "data"

# Names of the datasets
DIALECTS = "dialects"
LOWLIGHT = "lowlight"
REDDIT = "reddit"
MOVIES = "movies"
ESSAYS = "essays"
AEGIS = "aegis"
NSFW = "nsfw"


def get_data_home(path=None):
    """
    Return the path of the Construe data directory. This folder is used by
    dataset loaders to avoid downloading data several times.

    By default, this folder is colocated with the code in the install directory
    so that data shipped with the package can be easily located. Alternatively
    it can be set by the ``$CONSTRUE_DATA`` environment variable, or
    programmatically by giving a folder path. Note that the ``'~'`` symbol is
    expanded to the user home directory, and environment variables are also
    expanded when resolving the path.
    """
    if path is None:
        path = os.environ.get("CONSTRUE_DATA", DATADIR)

    path = os.path.expanduser(path)
    path = os.path.expandvars(path)

    if not os.path.exists(path):
        os.makedirs(path)

    return path


def find_dataset_path(dataset, data_home=None, fname=None, ext=None, raises=True):
    """
    Looks up the path to the dataset specified in the data home directory,
    which is found using the ``get_data_home`` function. By default data home
    is in a config directory in the user's home folder, but can be modified with the
    $CONSTRUE_DATA environment variable, or passing in a different directory.

    If the dataset is not found a ``DatasetsError`` is raised by default.
    """
    # Figure out the root directory of the datasets
    data_home = get_data_home(data_home)

    # Figure out the relative path to the dataset
    if fname is None:
        if ext is None:
            path = os.path.join(data_home, dataset)
        else:
            path = os.path.join(data_home, dataset, "{}{}".format(dataset, ext))
    else:
        path = os.path.join(data_home, dataset, fname)

    # Determine if the path exists
    if not os.path.exists(path):

        # Suppress exceptions if required
        if not raises:
            return None

        raise DatasetsError(
            f"could not find dataset at {path} - does it need to be downloaded?"
        )

    return path


def dataset_exists(dataset, data_home=None):
    """
    Checks to see if a directory with the name of the specified dataset exists
    in the data home directory, found with ``get_data_home``.
    """
    data_home = get_data_home(data_home)
    path = os.path.join(data_home, dataset)

    return os.path.exists(path) and os.path.isdir(path)


def dataset_archive(dataset, signature, data_home=None, ext=".zip"):
    """
    Checks to see if the dataset archive file exists in the data home directory,
    found with ``get_data_home``. By specifying the signature, this function
    also checks to see if the archive is the latest version by comparing the
    sha256sum of the local archive with the specified signature.
    """
    data_home = get_data_home(data_home)
    path = os.path.join(data_home, dataset + ext)

    if os.path.exists(path) and os.path.isfile(path):
        return sha256sum(path) == signature

    return False


def cleanup_dataset(dataset, data_home=None, ext=".zip"):
    """
    Removes the dataset directory and archive file from the data home directory.
    """
    removed = 0
    data_home = get_data_home(data_home)

    # Paths to remove
    datadir = os.path.join(data_home, dataset)
    archive = os.path.join(data_home, dataset + ext)

    # Remove directory and contents
    if os.path.exists(datadir):
        shutil.rmtree(datadir)
        removed += 1

    # Remove the archive file
    if os.path.exists(archive):
        os.remove(archive)
        removed += 1

    return removed



"""
Handle downloading datasets from our content URL
"""

from functools import partial

from .path import get_data_home
from .manifest import load_manifest
from ..cloud.download import download_zip
from .path import DIALECTS, LOWLIGHT, REDDIT, MOVIES, ESSAYS, AEGIS, NSFW

from ..exceptions import DatasetsError


def download_data(
    url, signature, data_home=None, replace=False, extract=True, progress=True
):
    """
    Downloads the zipped data set specified at the given URL, saving it to
    the data directory specified by ``get_data_home``. This function verifies
    the download with the given signature and extracts the archive.
    """
    data_home = get_data_home(data_home)
    download_zip(
        url, data_home, signature, replace=replace, extract=extract, progress=progress
    )


def _download_dataset(
    name, sample=True, data_home=True, replace=False, extract=True, progress=True
):
    """
    Downloads the zipped data set specified using the manifest URL, saving it to the
    data directory specified by ``get_data_home``. The download is verified with
    the given signature then extracted.
    """
    if sample and not name.endswith("-sample"):
        name = name + "-sample"

    datasets = load_manifest()
    if name not in datasets:
        raise DatasetsError(f"no dataset named {name} exists")

    info = datasets[name]
    kwargs = {
        "data_home": data_home,
        "replace": replace,
        "extract": extract,
        "progress": progress,
        "url": info["url"],
        "signature": info["signature"],
    }
    download_data(**kwargs)


download_dialects = partial(_download_dataset, DIALECTS)
download_lowlight = partial(_download_dataset, LOWLIGHT)
download_reddit = partial(_download_dataset, REDDIT)
download_movies = partial(_download_dataset, MOVIES)
download_essays = partial(_download_dataset, ESSAYS)
download_aegis = partial(_download_dataset, AEGIS)
download_nsfw = partial(_download_dataset, NSFW)


DOWNLOADERS = [
    download_dialects,
    download_lowlight,
    download_reddit,
    download_movies,
    download_essays,
    download_aegis,
    download_nsfw,
]


def download_all_datasets(
    sample=True, data_home=True, replace=True, extract=True, progress=True
):
    for f in DOWNLOADERS:
        f(
            sample=sample,
            data_home=data_home,
            replace=replace,
            extract=extract,
            progress=progress,
        )



"""
Utilities for model loading and conversion.
"""

import tensorflow as tf


class TFLiteGenerateModel(tf.Module):
    """
    Defines a model whose serving function is the generation call.
    """

    def __init__(self, model):
        super(TFLiteGenerateModel, self).__init__()
        self.model = model

    @tf.function(
        # shouldn't need static batch size, but throws exception without it (needs to be fixed)
        input_signature=[
            tf.TensorSpec((1, 80, 3000), tf.float32, name="input_features"),
        ],
    )
    def serving(self, input_features):
        outputs = self.model.generate(
            input_features,
            max_new_tokens=255,
            return_dict_in_generate=True,
        )
        return {"sequences": outputs["sequences"]}



"""
Managers for loading models
"""

import os
import shutil

from functools import partial

from .manifest import load_manifest
from .download import download_model
from ..exceptions import ModelsError
from .path import model_archive, cleanup_model
from .path import find_model_path, get_model_home
from .path import NSFW, LOWLIGHT, OFFENSIVE, GLINER
from .path import MOONDREAM, WHISPER, MOBILENET, MOBILEVIT

from tensorflow import lite as tflite
from transformers import WhisperProcessor


__all__ = [
    "load_all_models", "cleanup_all_models",
    "load_moondream", "cleanup_moondream",
    "load_whisper", "cleanup_whisper",
    "load_mobilenet", "cleanup_mobilenet",
    "load_mobilevit", "cleanup_mobilevit",
    "load_nsfw", "cleanup_nsfw",
    "load_lowlight", "cleanup_lowlight",
    "load_offensive", "cleanup_offensive",
    "load_gliner", "cleanup_gliner",
]


MODELS = load_manifest()


def _info(model):
    if model not in MODELS:
        raise ModelsError(f"no model named {model} exists")
    return MODELS[model]


def _model_path(name, tflite=True, model_home=None):
    info = _info(name)
    if not model_archive(name, info["signature"], model_home=model_home):
        # If the model does not exist, download and extract it
        kwargs = {
            "model_home": model_home, "replace": True, "extract": True,
            "url": info["url"], "signature": info["signature"]
        }
        download_model(**kwargs)

    if tflite:
        return find_model_path(name, model_home=model_home, ext=".tflite")
    return find_model_path(name, model_home=model_home)


def load_moondream(model_home=None):
    pass


def load_whisper(model_home=None):
    """
    Returns a tflite interpreter with the whisper model and the whisper prepocessor.
    """
    model_path = _model_path(WHISPER, model_home=model_home)
    proccessor_path = _model_path(WHISPER, tflite=False, model_home=model_home)

    model = tflite.Interpreter(model_path)
    processor = WhisperProcessor.from_pretrained(proccessor_path)
    return model, processor


def load_mobilenet(model_home=None):
    pass


def load_mobilevit(model_home=None):
    pass


def load_nsfw(model_home=None):
    pass


def load_lowlight(model_home=None):
    path = _model_path(LOWLIGHT, model_home=model_home)
    return tflite.Interpreter(path)


def load_offensive(model_home=None):
    pass


def load_gliner(model_home=None):
    pass


def load_all_models(model_home=None):
    """
    Load all available models as defined by __all__
    """
    models = {}
    module = globals()
    for name in __all__:
        if not name.startswith("load"):
            continue

        if name == "load_all_models":
            continue

        f = module[name]
        models[name] = f(model_home)

    return models


def cleanup_all_models(model_home=None):
    """
    Delete everything in the model home directory
    """
    with os.scandir(get_model_home(model_home)) as entries:
        for entry in entries:
            if entry.is_dir() and not entry.is_symlink():
                shutil.rmtree(entry.path)
            else:
                os.remove(entry.path)


cleanup_moondream = partial(cleanup_model, MOONDREAM)
cleanup_whisper = partial(cleanup_model, WHISPER)
cleanup_mobilenet = partial(cleanup_model, MOBILENET)
cleanup_mobilevit = partial(cleanup_model, MOBILEVIT)
cleanup_nsfw = partial(cleanup_model, NSFW)
cleanup_lowlight = partial(cleanup_model, LOWLIGHT)
cleanup_offensive = partial(cleanup_model, OFFENSIVE)
cleanup_gliner = partial(cleanup_model, GLINER)



"""
Defines the base class for all Benchmarks.
"""

import abc

from ..models import get_model_home
from ..datasets import get_data_home

from typing import Any, Generator, Dict, Union


class Benchmark(abc.ABC):
    """
    All benchmarks must subclass this class to ensure all properties and methods are
    correctly set for generic benchmarks to be run correctly.
    """

    @staticmethod
    @abc.abstractmethod
    def total(**kwargs):
        """
        For progress bar purposes should report the total number of instances in one
        run of the Benchmark. Generally this should be hard-coded but can also be
        computed if necessary.
        """
        pass

    def __init__(self, **kwargs):
        self._data_home = get_data_home(kwargs.pop("data_home", None))
        self._model_home = get_model_home(kwargs.pop("model_home", None))
        self._use_sample = kwargs.pop("use_sample", True)
        self._progress = kwargs.pop("progress", True)
        self._options = kwargs

    @property
    def data_home(self) -> str:
        if hasattr(self, "_data_home"):
            return self._data_home
        return get_data_home()

    @property
    def model_home(self) -> str:
        if hasattr(self, "_model_home"):
            return self._model_home
        return get_model_home()

    @property
    def use_sample(self) -> bool:
        return getattr(self, "_use_sample", True)

    @property
    def metadata(self) -> Dict:
        return getattr(self, "_metadata", None)

    @property
    def options(self) -> Union[Dict, None]:
        return getattr(self, "_options", {})

    @property
    @abc.abstractmethod
    def description(self):
        pass

    @abc.abstractmethod
    def before(self):
        """
        This method is called before the benchmark runs and should cause it to
        setup any datasets and models needed for the benchmark to run.
        """
        pass

    @abc.abstractmethod
    def after(self, cleanup: bool = True):
        """
        This method is called after the benchamrk is run; if cleanup is True the
        class should delete any cached datasets or models.
        """
        pass

    @abc.abstractmethod
    def instances(self, limit=None) -> Generator[Any, None, None]:
        """
        This method should yield all instances in the dataset at least once.
        """
        pass

    @abc.abstractmethod
    def preprocess(self, instance: Any) -> Any:
        """
        Any preprocessing that must be performed on an instance is handled with this
        method. This method is measured for latency and memory usage as well.
        """
        pass

    @abc.abstractmethod
    def inference(self, instance: Any) -> Any:
        """
        This represents the primary inference of the benchmark and is measured for
        latency and memory usage to add to the metrics.
        """
        pass



"""
Moondream is a computer vision model (image to text) that is optimized for use
on embedded devices and serves as an example model in content moderation use
cases where the image is captioned and then the caption is moderated.
"""

from .exceptions import DatasetsError
from .benchmark import Benchmark, limit_generator
from .models import load_moondream, cleanup_moondream
from .datasets import load_nsfw, cleanup_nsfw, DATASETS


class MoonDream(Benchmark):

    @staticmethod
    def total(**kwargs):
        # Return the number of nsfw images from the manifest
        use_sample = kwargs.pop("use_sample", True)
        name = "nsfw-sample" if use_sample else "nsfw"
        if name not in DATASETS:
            raise DatasetsError("nsfw dataset not found in manifest")
        return DATASETS[name]["instances"]

    @property
    def description(self):
        return (
            "performs content moderation by captioning an image and "
            "then using the offensive speech model to moderate the caption"
        )

    def before(self):
        model, processor = load_moondream(model_home=self.model_home)
        self.model = model
        self.processor = processor
        self.generate = self.model.get_signature_runner()

    def after(self, cleanup=True):
        if cleanup:
            cleanup_moondream(model_home=self.model_home)
            cleanup_nsfw(data_home=self.data_home, sample=self.use_sample)

    def instances(self, limit=None):
        dataset = load_nsfw(data_home=self.data_home, sample=self.use_sample)
        return limit_generator(dataset, limit)

    def preprocess(self, instance):
        raise NotImplementedError("MoonDream preprocess not implemented")

    def inference(self, instance):
        raise NotImplementedError("MoonDream inference not implemented")



"""
Manifest handlers for downloading cloud resources and checking signatures.
"""

import os
import json
import glob
import zipfile

from urllib.parse import urljoin

from .signature import sha256sum
from ..version import get_version


BUCKET = "construe"
BASE_URL = "https://storage.googleapis.com/"

MODELS = "models"
DATASETS = "datasets"


def load_manifest(path):
    with open(path, "r") as f:
        return json.load(f)


def generate_manifest(fixtures, out, upload_type, extra=None):
    manifest = {}
    version = get_version(short=True)

    # Sort the list of paths by name
    paths = list(glob.glob(os.path.join(fixtures, "*.zip")))
    paths.sort()

    for path in paths:
        fname = os.path.basename(path)
        name, _ = os.path.splitext(fname)

        manifest[name] = {
            "url": make_fixture_url(fname, upload_type=upload_type, version=version),
            "signature": sha256sum(path),
            "size": {
                "compressed": os.path.getsize(path),
                "decompressed": get_uncompressed_size(path),
            },
        }

        if extra is not None:
            if callable(extra):
                manifest[name].update(extra(path=path, name=name, **manifest[name]))
            else:
                manifest[name].update(extra)

    with open(out, "w") as o:
        json.dump(manifest, o, indent=2)


def make_fixture_url(fname, upload_type, version=None):
    # Bucket must be joined here and not make_fixture_path to support uploading
    path = make_fixture_path(fname, upload_type, version)
    path = os.path.join(BUCKET, path)
    return urljoin(BASE_URL, path)


def make_fixture_path(fname, upload_type, version=None):
    version = version or get_version(short=True)
    return os.path.join(f"v{version}", upload_type, fname)


def get_uncompressed_size(path: str) -> int:
    bytes = 0
    with zipfile.ZipFile(path, 'r') as zf:
        for info in zf.infolist():
            bytes += info.file_size
    return bytes



"""
The cloud module helps developers create manifests and upload datasets and manifests
to cloud storage where they're cached for use with the LLM benchmark utility.

This module should not be used by benchmark users; use the datasets and models
utility directly instead.
"""



"""
Tests the data set loaders and downloaders to make sure they work as expected.
"""

import os
import pytest

from construe.datasets.loaders import *


# Data loader tests required downloading data from the bucket; skipping them by default
# unless the environment variable is set will make the tests run faster. In CI this
# environment variable can be set to protect from regressions.
CONSTRUE_TEST_DATA_LOADERS = os.environ.get("CONSTRUE_TEST_DATA_LOADERS", False)
CONSTRUE_TEST_SAMPLE_LOADERS = os.environ.get("CONSTRUE_TEST_SAMPLE_LOADERS", False)


@pytest.mark.skipif(not CONSTRUE_TEST_DATA_LOADERS, reason="skipping data loaders tests")
def test_load_dialects(tmpdir):
    count = 0
    for dialect in load_dialects(data_home=tmpdir, sample=False):
        count += 1
        assert isinstance(dialect, str) and dialect.endswith(".wav")

    assert count == 17877
    cleanup_dialects(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_SAMPLE_LOADERS, reason="skipping data loaders tests")
def test_load_dialects_sample(tmpdir):
    count = 0
    for dialect in load_dialects(data_home=tmpdir, sample=True):
        count += 1
        assert isinstance(dialect, str) and dialect.endswith(".wav")

    assert count == 1785
    cleanup_dialects(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_DATA_LOADERS, reason="skipping data loaders tests")
def test_load_lowlight(tmpdir):
    count = 0
    for image in load_lowlight(data_home=tmpdir, sample=False):
        count += 1
        assert isinstance(image, str) and image.endswith(".png")

    assert count == 1000
    cleanup_dialects(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_SAMPLE_LOADERS, reason="skipping data loaders tests")
def test_load_lowlight_sample(tmpdir):
    count = 0
    for image in load_lowlight(data_home=tmpdir, sample=True):
        count += 1
        assert isinstance(image, str) and image.endswith(".png")

    assert count == 475
    cleanup_lowlight(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_DATA_LOADERS, reason="skipping data loaders tests")
def test_load_reddit(tmpdir):
    count = 0
    for row in load_reddit(data_home=tmpdir, sample=False):
        count += 1
        assert isinstance(row, dict) and "comment" in row

    assert count == 3844
    cleanup_reddit(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_SAMPLE_LOADERS, reason="skipping data loaders tests")
def test_load_reddit_sample(tmpdir):
    count = 0
    for row in load_reddit(data_home=tmpdir, sample=True):
        count += 1
        assert isinstance(row, dict) and "comment" in row

    assert count == 957
    cleanup_reddit(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_DATA_LOADERS, reason="skipping data loaders tests")
def test_load_movies(tmpdir):
    count = 0
    for image in load_movies(data_home=tmpdir, sample=False):
        count += 1
        assert isinstance(image, str) and image.endswith(".jpg")

    assert count == 106844
    cleanup_movies(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_SAMPLE_LOADERS, reason="skipping data loaders tests")
def test_load_movies_sample(tmpdir):
    count = 0
    for image in load_movies(data_home=tmpdir, sample=True):
        count += 1
        assert isinstance(image, str) and image.endswith(".jpg")

    assert count == 5465
    cleanup_movies(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_DATA_LOADERS, reason="skipping data loaders tests")
def test_load_essays(tmpdir):
    count = 0
    for row in load_essays(data_home=tmpdir, sample=False):
        count += 1
        assert isinstance(row, dict) and "essay" in row

    assert count == 2078
    cleanup_essays(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_SAMPLE_LOADERS, reason="skipping data loaders tests")
def test_load_essays_sample(tmpdir):
    count = 0
    for row in load_essays(data_home=tmpdir, sample=True):
        count += 1
        assert isinstance(row, dict) and "essay" in row

    assert count == 512
    cleanup_essays(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_DATA_LOADERS, reason="skipping data loaders tests")
def test_load_aegis(tmpdir):
    count = 0
    for row in load_aegis(data_home=tmpdir, sample=False):
        count += 1
        assert isinstance(row, dict) and "labels_0" in row and "text" in row

    assert count == 11997
    cleanup_aegis(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_SAMPLE_LOADERS, reason="skipping data loaders tests")
def test_load_aegis_sample(tmpdir):
    count = 0
    for row in load_aegis(data_home=tmpdir, sample=True):
        count += 1
        assert isinstance(row, dict) and "labels_0" in row and "text" in row

    assert count == 3030
    cleanup_aegis(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_DATA_LOADERS, reason="skipping data loaders tests")
def test_load_nsfw(tmpdir):
    count = 0
    for image in load_nsfw(data_home=tmpdir, sample=False):
        count += 1
        assert isinstance(image, str) and image.endswith(".jpg")

    assert count == 215
    cleanup_nsfw(data_home=tmpdir)


@pytest.mark.skipif(not CONSTRUE_TEST_SAMPLE_LOADERS, reason="skipping data loaders tests")
def test_load_nsfw_sample(tmpdir):
    count = 0
    for image in load_nsfw(data_home=tmpdir, sample=True):
        count += 1
        assert isinstance(image, str) and image.endswith(".jpg")

    assert count == 53
    cleanup_nsfw(data_home=tmpdir)



"""
Tests for the basic benchmarking module.
"""

import torch

from construe.basic import *


def test_batched_dot():
    # Input for benchmarking
    x = torch.randn(10000, 64)

    # Ensure that both functions compute the same output
    assert batched_dot_mul_sum(x, x).allclose(batched_dot_bmm(x, x))



"""
Handles limiting the output of generators.
"""

from typing import Generator, Optional


def limit_generator(generator: Generator, limit: Optional[int] = None) -> Generator:
    """
    Limit the output of a generator to a certain number of items.
    """
    if limit is None:
        yield from generator
        return

    for i, item in enumerate(generator):
        if i >= limit:
            return
        yield item



"""
Testing for the metrics module.
"""

import pytest

from construe.metrics import Metric, Measurement
from construe.metrics import select_duration_unit


class TestMetric(object):
    """
    Metric dataclass tests
    """

    @pytest.mark.parametrize(
        "metric,expected",
        [
            (Metric(), "Metric"),
            (Metric(label="Foo"), "Foo"),
            (Metric(label="Foo", sub_label="Bar"), "Foo: Bar"),
            (Metric(env="MediaTek"), "Metric for MediaTek"),
            (Metric(env="MediaTek", device="tflite"), "Metric for MediaTek on tflite"),
        ],
    )
    def test_title(self, metric, expected):
        """
        Test title rendering
        """
        assert metric.title == expected


@pytest.mark.usefixtures("completed_benchmark")
class TestMeasurement(object):
    """
    Measurement dataclass tests
    """

    def test_per_run_default(self):
        """
        Assert that per_run is one by default
        """
        m = Measurement(metric=Metric(), raw_metrics=[1, 2, 3])
        assert m.per_run == 1, "per run is not set to one!"

    def test_stats(self):
        """
        Test stats calculations of measurement
        """
        assert self.measurement.median == 23.260444076123438
        assert self.measurement.mean == 23.112752271750523
        assert self.measurement.iqr == 8.44631785879561


@pytest.mark.parametrize(
    "t,expected",
    [
        (1.000, ("s", 1)),
        (5.000, ("s", 1)),
        (10.000, ("s", 1)),
        (50.000, ("s", 1)),
        (100.000, ("s", 1)),
        (500.000, ("s", 1)),
        (0.1, ("ms", 1e-3)),
        (0.6, ("ms", 1e-3)),
        (0.01, ("ms", 1e-3)),
        (0.03, ("ms", 1e-3)),
        (0.001, ("ms", 1e-3)),
        (0.004, ("ms", 1e-3)),
        (0.0001, ("us", 1e-6)),
        (0.0008, ("us", 1e-6)),
        (0.00001, ("us", 1e-6)),
        (0.00002, ("us", 1e-6)),
        (0.000001, ("us", 1e-6)),
        (0.000007, ("us", 1e-6)),
        (0.0000001, ("ns", 1e-9)),
        (0.0000009, ("ns", 1e-9)),
        (0.00000001, ("ns", 1e-9)),
        (0.00000005, ("ns", 1e-9)),
        (0.000000001, ("ns", 1e-9)),
        (0.000000003, ("ns", 1e-9)),
    ],
)
def test_select_duration_unit(t, expected):
    assert select_duration_unit(t) == expected



"""
Testing for serialize module.
"""

import json

from construe.metrics import dumps
from construe.metrics import Metric, Measurement


def test_serialize_string():
    measurement = Measurement(
        raw_metrics=[
            21.82307791099665,
            26.531772732839926,
            27.505767531015223,
            16.403111227499263,
            28.29817512637112,
            16.76156931690454,
            25.319997297409323,
        ],
        metric=Metric(
            label="Basic",
            sub_label="mul/sum",
            description="Basic benchmark for matrix operations",
            env="MediaTek Genio EVK 1200",
            device="TFLite APU",
        ),
        metadata={"testing": True},
    )

    out = dumps(measurement)
    assert len(out) > 20, "no data was output from the dump method"

    data = json.loads(out)
    assert "type" in data, "no type was in the data struct"
    assert data["type"] == Measurement.__name__, "incorrect type added to data"



import pytest
from construe.benchmark.limit import limit_generator

"""
Test the benchmark generator helpers.
"""


@pytest.mark.parametrize(
    "limit,expected",
    [
        (None, 10),
        (10, 10),
        (15, 10),
        (5, 5),
        (1, 1),
        (0, 0),
        (-10, 0),
    ],
)
def test_limit_generator(limit, expected):
    def sample_generator():
        for i in range(10):
            yield i

    gen = limit_generator(sample_generator(), limit=limit)
    assert len(list(gen)) == expected



"""
Offensive speech benchmark runner
"""

from .exceptions import DatasetsError
from .benchmark import Benchmark, limit_generator
from .models import load_offensive, cleanup_offensive
from .datasets import load_aegis, cleanup_aegis, DATASETS


class Offensive(Benchmark):

    @staticmethod
    def total(**kwargs):
        ## Return the number of aegis comments from the manifest
        use_sample = kwargs.pop("use_sample", True)
        name = "aegis-sample" if use_sample else "aegis"
        if name not in DATASETS:
            raise DatasetsError("aegis dataset not found in manifest")
        return DATASETS[name]["instances"]

    @property
    def description(self):
        return (
            "applies the offensive speech detection model "
            "to the aegis content safety dataset"
        )

    def before(self):
        model, processor = load_offensive(model_home=self.model_home)
        self.model = model
        self.processor = processor

    def after(self, cleanup=True):
        if cleanup:
            cleanup_offensive(model_home=self.model_home)
            cleanup_aegis(data_home=self.data_home, sample=self.use_sample)

    def instances(self, limit=None):
        dataset = load_aegis(data_home=self.data_home, sample=self.use_sample)
        return limit_generator(dataset, limit)

    def preprocess(self, instance):
        raise NotImplementedError("Offensive preprocess not implemented")

    def inference(self, instance):
        raise NotImplementedError("Offensive inference not implemented")



"""
Handle downloading models from content URLs
"""

from functools import partial

from .path import get_model_home
from .manifest import load_manifest
from ..cloud.download import download_zip
from .path import NSFW, LOWLIGHT, OFFENSIVE, GLINER
from .path import MOONDREAM, WHISPER, MOBILENET, MOBILEVIT

from ..exceptions import ModelsError


def download_model(
    url, signature, model_home=None, replace=False, extract=True, progress=True
):
    """
    Downloads the zipped model file specified at the given URL saving it to the models
    directory specified by ``get_model_home``. The download is verified with the
    given signature then extracted.
    """
    model_home = get_model_home(model_home)
    download_zip(
        url, model_home, signature, replace=replace, extract=extract, progress=progress
    )


def _download_model(name, model_home=None, replace=False, extract=True, progress=True):
    """
    Downloads the zipped model file specified using the manifest URL, saving it to the
    models directory specified by ``get_model_home``. The download is verified with
    the given signature then extracted.
    """
    models = load_manifest()
    if name not in models:
        raise ModelsError(f"no model named {name} exists")

    info = models[name]
    kwargs = {
        "model_home": model_home,
        "replace": replace,
        "extract": extract,
        "progress": progress,
        "url": info["url"],
        "signature": info["signature"],
    }
    download_model(**kwargs)


download_moondream = partial(_download_model, MOONDREAM)
download_whisper = partial(_download_model, WHISPER)
download_mobilenet = partial(_download_model, MOBILENET)
download_mobilevit = partial(_download_model, MOBILEVIT)
download_nsfw = partial(_download_model, NSFW)
download_lowlight = partial(_download_model, LOWLIGHT)
download_offensive = partial(_download_model, OFFENSIVE)
download_gliner = partial(_download_model, GLINER)


DOWNLOADERS = [
    download_moondream,
    download_whisper,
    download_mobilenet,
    download_mobilevit,
    download_nsfw,
    download_lowlight,
    download_offensive,
    download_gliner,
]


def download_all_models(model_home=None, replace=True, extract=True, progress=True):
    for f in DOWNLOADERS:
        f(model_home=model_home, replace=replace, extract=extract, progress=progress)



"""
LowLight image enhancement benchmark runner
"""

import os
import numpy as np
import tensorflow as tf

from .datasets import DATASETS
from .exceptions import DatasetsError
from .benchmark import Benchmark, limit_generator
from .models import load_lowlight as load_lowlight_model
from .models import cleanup_lowlight as cleanup_lowlight_model
from .datasets import load_lowlight as load_lowlight_dataset
from .datasets import cleanup_lowlight as cleanup_lowlight_dataset


class LowLight(Benchmark):

    @staticmethod
    def total(**kwargs):
        ## Return number of lowlight images from the manifest
        use_sample = kwargs.pop("use_sample", True)
        name = "lowlight-sample" if use_sample else "lowlight"
        if name not in DATASETS:
            raise DatasetsError("lowlight dataset not found in manifest")
        return DATASETS[name]["classes"]["low"]

    @property
    def description(self):
        return (
            "lowlight enhances image quality using a convolutional "
            "model that understands how to enrich low-light images"
        )

    def before(self):
        # Load and setup the interpreter for the lowlight dataset
        model = load_lowlight_model(model_home=self.model_home)
        model.resize_tensor_input(0, [1, 400, 600, 3])
        model.allocate_tensors()

        self.model = model
        self.input_details = model.get_input_details()
        self.output_details = model.get_output_details()

    def after(self, cleanup=True):
        if cleanup:
            cleanup_lowlight_model(model_home=self.model_home)
            cleanup_lowlight_dataset(data_home=self.data_home, sample=self.use_sample)

    def instances(self, limit=None):
        dataset = load_lowlight_dataset(
            data_home=self.data_home, sample=self.use_sample,
        )

        def filter_instances(dataset):
            for instance in dataset:
                if os.path.dirname(instance).endswith("low"):
                    yield instance

        return limit_generator(filter_instances(dataset), limit)

    def preprocess(self, instance):
        image = tf.io.read_file(instance)
        image = tf.image.decode_png(image, channels=3)
        image.set_shape([None, None, 3])
        image = tf.cast(image, dtype=tf.float32) / 255.0
        image = np.expand_dims(image, axis=0)

        return image

    def inference(self, instance):
        self.model.set_tensor(self.input_details[0]["index"], instance)
        self.model.invoke()
        self.model.get_tensor(self.output_details[0]["index"])
